/*
 * thrulay.c -- network throughput tester (the client part).
 * 
 * Written by Stanislav Shalunov, http://www.internet2.edu/~shalunov/
 *            Bernhard Lutzmann, belu@users.sf.net
 * 
 * Copyright 2003, Internet2.
 * Legal conditions are in file LICENSE
 * (MD5 = ecfa50d1b0bfbb81b658c810d0476a52).
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef HAVE_STDINT_H
#include <stdint.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <limits.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>	/* need this under FreeBSD for
				 * <netinet/ip.h> */
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <assert.h>
#include <signal.h>
#include <sys/param.h>
#include <sys/uio.h>
#include <time.h>
#include <sys/time.h>
#include <unistd.h>
#include <string.h>
#include <strings.h>
#include <math.h>
#include <errno.h>
#include "thrulay.h"
#include "util.h"
#include "rcs.h"

RCS_ID("@(#) $Id: thrulay.c,v 1.3 2005/10/17 23:28:43 shalunov Exp $")

#define THRULAY_VERSION		"thrulay/2"
#define THRULAY_GREET		THRULAY_VERSION "+"

#define IPV4_HEADER_SIZE	20
#define IPV6_HEADER_SIZE	40
#define UDP_HEADER_SIZE		8
#define UDP_PAYLOAD_SIZE	24

#define UDP_PORT		5003
#define TRY_UDP_PORTS		1000

#ifndef SOL_IP
#ifdef IPPROTO_IP
#define SOL_IP			IPPROTO_IP
#endif
#endif

#ifndef SOL_TCP
#ifdef IPPROTO_TCP
#define SOL_TCP			IPPROTO_TCP
#endif
#endif

/* IP_MTU not defined in <bits/in.h> under Linux */
#if defined (__LINUX__) && !defined(IP_MTU)
#define IP_MTU 14
#endif

#if defined (__SOLARIS__)
#define RANDOM_MAX		4294967295UL	/* 2**32-1 */
#elif defined (__DARWIN__)
#define RANDOM_MAX		LONG_MAX	/* Darwin */
#else
#define RANDOM_MAX		RAND_MAX	/* Linux, FreeBSD */
#endif

fd_set rfds_orig, wfds_orig;
int maxfd = 0;
char *block = NULL;
int stop_test = 0;
int server_block_size;
int local_window, server_window;
int mtu, mss;

/* UDP test */
unsigned int client_port, server_port;
unsigned int packet_size;
unsigned int protocol_rate;	/* In packets per 1000 seconds. */
int tcp_sock, udp_sock;
uint64_t npackets;
struct sockaddr *server = NULL;
socklen_t server_len;

/* Set default options. */
void
options_init (void)
{
	opt.server_name = NULL;
	opt.num_streams = 1;
	opt.test_duration = 60;
	opt.reporting_interval = 1;
	opt.window = 4194304;
	opt.block_size = 0;
	opt.port = 5003;
	opt.rate = 0;
	opt.dscp = 0;
	opt.busywait = 0;
}

void
tcp_stats_init (void)
{
	int id;

	for (id = 0; id < opt.num_streams; id++)
	{
		/* Blocks */
		stats[id].blocks_since_first = 0;
		stats[id].blocks_since_last = 0;

		/* RTT */
		stats[id].min_rtt_since_first = 1000.0;
		stats[id].min_rtt_since_last = 1000.0;
		stats[id].max_rtt_since_first = -1000.0;
		stats[id].max_rtt_since_last = -1000.0;
		stats[id].tot_rtt_since_first = 0.0;
		stats[id].tot_rtt_since_last = 0.0;
	}
}

static void __attribute__((noreturn))
usage(void)
{
	fprintf(stderr, "Usage: thrulay [-b] [-D DSCP] [-m#] [-t#] [-i#] "
			"[-w#] [-l#] [-p#] [-u#] host\n");
	fprintf(stderr, "\t-b\t\tBusy wait in UDP test\n");
	fprintf(stderr, "\t-D DSCP\t\tDSCP value for TOS byte "
			"(default: unset)\n");
	fprintf(stderr, "\t-m#\t\tnumber of test streams (default: 1)\n");
	fprintf(stderr, "\t-t#\t\ttest duration, in seconds (default: 60s)\n");
	fprintf(stderr, "\t-i#\t\treporting interval, in seconds "
			"(default: 1s)\n");
	fprintf(stderr, "\t-w#\t\twindow, in bytes (default: 4194304B)\n");
	fprintf(stderr, "\t-l#\t\tblock size (default: 8192B)\n");
	fprintf(stderr, "\t-p#\t\tserver port (default: 5003)\n");
	fprintf(stderr, "\t-u#[kMGT]\tUDP mode with given rate "
				"(default: off)\n"
			"\t\t\tIn UDP mode, rate is in bits per second and "
				"can be\n"
			"\t\t\tfollowed by a SI suffix (k for 1000, M for "
				"100000);\n"
			"\t\t\tdefault packet size, 1500, can be changed with "
				"-l;\n"
			"\t\t\twindow size becomes the UDP send buffer size;\n"
			"\t\t\treporting interval is ignored.\n");
	fprintf(stderr, "\thost\t\tserver to send test data to (no default)\n");
	exit(1);
}

/* Process new timestamp. */
void
new_timestamp(int id, struct timeval *tv)
{
	struct timeval this;
	double relative;

	if (tsc_gettimeofday(&this) == -1) {
		perror("gettimeofday");
		error(ERR_FATAL, "Unable to process new timestamp.");
	}
	normalize_tv(&this);

	relative = time_diff(tv, &this);
	if (relative < 0) {
		error(ERR_WARNING, "negative delay, ignoring block");
		return;
	}

	/* Update statistics for stream. */
	stats[id].blocks_since_first++;
	stats[id].blocks_since_last++;
	if (stats[id].min_rtt_since_first > relative)
		stats[id].min_rtt_since_first = relative;
	if (stats[id].min_rtt_since_last > relative)
		stats[id].min_rtt_since_last = relative;
	if (stats[id].max_rtt_since_first < relative)
		stats[id].max_rtt_since_first = relative;
	if (stats[id].max_rtt_since_last < relative)
		stats[id].max_rtt_since_last = relative;
	stats[id].tot_rtt_since_first += relative;
	stats[id].tot_rtt_since_last += relative;
}

/* If test duration is over, stops TCP test. */
void
timer_end (struct timeval *now)
{
	if (now->tv_sec > timer.stop.tv_sec
			|| (now->tv_sec == timer.stop.tv_sec
				&& now->tv_usec >= timer.stop.tv_usec)) {
		thrulay_tcp_stop();
	}
}

/* If progress report should be displayed does so and updates timer values. */
void
timer_report (struct timeval *now)
{
	if (now->tv_sec > timer.next.tv_sec
			|| (now->tv_sec == timer.next.tv_sec
				&& now->tv_usec >= timer.next.tv_usec)) {
		thrulay_tcp_report();

		timer.last.tv_sec = now->tv_sec;
		timer.last.tv_usec = now->tv_usec;

		while (timer.next.tv_sec < now->tv_sec
				|| (timer.next.tv_sec == now->tv_sec
					&& timer.next.tv_usec <= now->tv_usec))
			timer.next.tv_sec += opt.reporting_interval;
	}
}

void
timer_check (void)
{
	struct timeval this;

	if (tsc_gettimeofday(&this) == -1) {
		perror("gettimeofday");
		error(ERR_FATAL, "gettimeofday(): failed");
	}
	normalize_tv(&this);

	timer_report(&this);
	timer_end(&this);
}

/* Stop timer. Calculates runtime and saves this to `timer.runtime'. */
void
timer_stop (void)
{
	struct timeval this;

	if (tsc_gettimeofday(&this) == -1) {
		perror("gettimeofday");
		error(ERR_FATAL, "gettimeofday(): failed");
	}
	normalize_tv(&this);

	/* Set final runtime. */
	timer.runtime = time_diff(&timer.start, &this);
}

/* Start timer. This saves the starting time to `timer.start', saves time when
 * next progress report should be displayed in `timer.next' and time when
 * test should stop in `timer.stop'. This are all timeval structures. */
void
timer_start (void)
{
	if (tsc_gettimeofday(&timer.start) == -1) {
		perror("gettimeofday");
		error(ERR_FATAL, "Unable to start timer.");
	}
	normalize_tv(&timer.start);

	timer.last.tv_sec = timer.start.tv_sec;
	timer.last.tv_usec = timer.start.tv_usec;

	timer.next.tv_sec = timer.start.tv_sec + opt.reporting_interval;
	timer.next.tv_usec = timer.start.tv_usec;

	timer.stop.tv_sec = timer.start.tv_sec + opt.test_duration;
	timer.stop.tv_usec = timer.start.tv_usec;
}

void
thrulay_tcp_report_final_id (int id)
{
	if (stats[id].blocks_since_first == 0) {
		/* This stream was not very active :) */
		printf("#(%2d) %8.3f %8.3f %8.3f %8.3f %8.3f %8.3f\n",
				id, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0);
		return;
	}

	printf("#(%2d) %8.3f %8.3f %8.3f %8.3f %8.3f %8.3f\n",
		id,
		0.0,
		timer.runtime,
		(double)stats[id].blocks_since_first *
			(double)server_block_size * 8.0 / 1000000.0 /
			timer.runtime,
		stats[id].min_rtt_since_first * 1000.0,
		stats[id].tot_rtt_since_first * 1000.0 /
			(double)stats[id].blocks_since_first,
		stats[id].max_rtt_since_first * 1000.0);
}

void
thrulay_tcp_report_final (void)
{
	double mbs = 0.0;
	double min_rtt = 1000.0;
	double max_rtt = -1000.0;
	double avg_rtt_sum = 0.0;
	int id;

	for (id = 0; id < opt.num_streams; id++)
		thrulay_tcp_report_final_id(id);

	/* Now calculate global statistics. */
	for (id = 0; id < opt.num_streams; id++) {
		/* Calculate throughput of all streams together. */
		mbs += (double)stats[id].blocks_since_first *
			(double)server_block_size * 8.0 / 1000000.0 /
			timer.runtime;

		/* Calculate minimum RTT */
		min_rtt = (min_rtt < stats[id].min_rtt_since_first ?
				min_rtt : stats[id].min_rtt_since_first);

		/* Calculate maximum RTT */
		max_rtt = (stats[id].max_rtt_since_first < max_rtt ?
				max_rtt : stats[id].max_rtt_since_first);

		/* Calculate sum of average RTT */
		if (stats[id].blocks_since_first != 0) {
			avg_rtt_sum += stats[id].tot_rtt_since_first * 1000.0 /
				(double)stats[id].blocks_since_first;
		}
	}

	printf("#(**) %8.3f %8.3f %8.3f %8.3f %8.3f %8.3f\n",
			0.0,
			timer.runtime,			/* global runtime */
			mbs,				/* MB/s */
			min_rtt * 1000.0,		/* minimal RTT */
			avg_rtt_sum / opt.num_streams,	/* average RTT */
			max_rtt * 1000.0);		/* maximal RTT */
}

/* Variables for progress reporting. */
char report_buffer[STREAMS_MAX * 80];
char *report_buffer_ptr = NULL;
int report_buffer_len = 0;

void
thrulay_tcp_report_id (int id)
{
	struct timeval this;
	double diff_first_last;
	double relative;	/* Time difference between now and last */
	int n;

	if (tsc_gettimeofday(&this) == -1) {
		perror("gettimeofday");
		error(ERR_FATAL, "Unable to print stream stats.");
	}
	normalize_tv(&this);

	diff_first_last = time_diff(&timer.start, &timer.last);
	relative = time_diff(&timer.last, &this);

	/* Time must be monotonically increasing, at least
	 * roughly, for this program to work. */

	if (stats[id].blocks_since_last == 0) {
		n = sprintf(report_buffer_ptr,
				" (%2d) %8.3f %8.3f %8.3f %8.3f %8.3f %8.3f\n",
				id,
				diff_first_last,
				diff_first_last + relative,
				0.0,	/* MB/s */
				0.0,	/* min RTT since last */
				0.0,	/* avg RTT since last */
				0.0);	/* max RTT since last */
	} else {
		n = sprintf(report_buffer_ptr,
				" (%2d) %8.3f %8.3f %8.3f %8.3f %8.3f %8.3f\n",
				id,
				diff_first_last,
				diff_first_last + relative,
				(double)stats[id].blocks_since_last *
					(double)server_block_size *
					8.0 / 1000000.0 / relative,
				stats[id].min_rtt_since_last*1000.0,
				stats[id].tot_rtt_since_last*1000.0 /
					(double)stats[id].blocks_since_last,
				stats[id].max_rtt_since_last*1000.0);
	}
	report_buffer_ptr += n;
	report_buffer_len += n;

	stats[id].blocks_since_last = 0;
	stats[id].min_rtt_since_last = 1000.0;
	stats[id].max_rtt_since_last = -1000.0;
	stats[id].tot_rtt_since_last = 0.0;
}

void
thrulay_tcp_report (void)
{
	int id;

	report_buffer_ptr = &report_buffer[0];
	report_buffer_len = 0;

	for (id = 0; id < opt.num_streams; id++)
		thrulay_tcp_report_id(id);

	/* Display progress report. */
	write_exactly(STDOUT_FILENO, report_buffer, report_buffer_len);
}

/*
 * Return a socket connected to the right TCP port on the right
 * server.  Use the given window size, if non-zero, locally.  Store
 * the actual window size in *real_window.  Store server socket
 * address structure in `saptr' (if both saptr and lenp are non-NULL) and
 * size of socket address structure in `lenp'.
 */
int
name2socket(char *server_name, int port, int window, int *real_window,
	void **saptr, socklen_t *lenp)
{
	int sockfd, n;
	struct addrinfo hints, *res, *ressave;
	char service[7];

	bzero(&hints, sizeof(struct addrinfo));
	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;

	snprintf(service, sizeof(service), "%d", port);

	if ((n = getaddrinfo(server_name, service, &hints, &res)) != 0) {
		fprintf(stderr, "getaddrinfo(): %s\n", gai_strerror(n));
		error(ERR_FATAL, "getaddrinfo(): failed");
		/* NOTREACHED */
	}
	ressave = res;

	do {
		sockfd = socket(res->ai_family, res->ai_socktype,
				res->ai_protocol);
		if (sockfd < 0)
			continue;	/* ignore this one */

		if (window)
			*real_window = set_window_size(sockfd, window);

		if (connect(sockfd, res->ai_addr, res->ai_addrlen) == 0)
			break;		/* success */

		close(sockfd);
	} while ((res = res->ai_next) != NULL);

	if (res == NULL) {
		error(ERR_FATAL, "could not establish connection to server");
		/* NOTREACHED */
	}

	if (saptr && lenp) {
		*saptr = malloc(res->ai_addrlen);
		if (*saptr == NULL) {
			perror("malloc");
			error(ERR_FATAL, "malloc(): failed");
			/* NOTREACHED */
		}
		memcpy(*saptr, res->ai_addr, res->ai_addrlen);
		*lenp = res->ai_addrlen;
	}

	freeaddrinfo(ressave);

	return sockfd;
}

/* Read the thrulay greeting from a TCP connection associated with socket s. */
void
read_greeting(int s)
{
	char buf[1024];
	int rc;
	size_t greetlen = sizeof(THRULAY_GREET) - 1;

	rc = read_exactly(s, buf, greetlen);
	assert(rc <= (int) greetlen);
	if (rc != (int) greetlen) {
		if (rc == -1)
			perror("read");
		error(ERR_FATAL, "could not read server banner");
		/* NOTREACHED */
	}
	if (strncmp(buf, THRULAY_VERSION, sizeof(THRULAY_VERSION) - 1) != 0)
		error(ERR_FATAL, "not a thrulay server responded");
	if (buf[greetlen - 1] != '+') {
		error(ERR_WARNING, "connection rejected");
		rc = read(s, buf, sizeof(buf) - 1);
		buf[sizeof(buf) - 1] = '\0';
		if (rc == -1) {
			perror("reading rejection reason");
			error(ERR_FATAL, "could not read rejection reason");
			/* NOTREACHED */
		}
		assert(rc < (int) sizeof(buf));
		buf[rc] = '\0';
		fprintf(stderr, "server said: %s", buf);
		if (buf[rc-1] != '\n')
			fprintf(stderr, "\n");
		error(ERR_FATAL, "stop");
		/* NOTREACHED */
	}
}

void
write_proposal(int s, char *proposal, int proposal_size)
{
	int rc;

	rc = write_exactly(s, proposal, (size_t) proposal_size);
	assert(rc <= proposal_size);
	if (rc < proposal_size) {
		if (rc == -1)
			perror("write");
		error(ERR_FATAL, "could not write session proposal");
		/* NOTREACHED */
	}
}

/* Read response to a proposal.  Return the size of the response. */
int
read_response(int s, char *buf, int max)
{
	int rc;

	/* XXX: Assume that few-byte session response will come in one
           TCP packet and read in one block. */
	rc = read(s, buf, max - 1);
	assert(rc < max);
	if (rc == -1) {
		perror("read");
		error(ERR_FATAL, "could not read session response");
		/* NOTREACHED */
	} else if (rc == 0) {
		error(ERR_FATAL, "server closed connection after proposal");
		/* NOTREACHED */
	}
	assert(rc > 0);
	buf[rc] = '\0';
	return rc;
}

void
thrulay_udp_report_final (void)
{
	int rc;
	char buf[65536];

	snprintf(buf, sizeof(buf), "+%llu:", (long long unsigned)npackets);

	rc = write_exactly(tcp_sock, buf, strlen(buf));
	if (rc == -1)
		error(ERR_FATAL, "could not write terminating message");

	while ((rc = read(tcp_sock, buf, sizeof(buf) - 1)) != 0) {
		if (rc == -1) {
			perror("read");
			error(ERR_FATAL, "could not read session results");
		}
		write_exactly(STDOUT_FILENO, buf, rc);
	}
}

void
thrulay_udp_start (void)
{
	int rc;
	int val;
	char buf[65536];
	char random_state[256];
	int to_write;
	uint64_t packet;
	uint32_t msb, lsb;
	char nonce[8];
	int n, response_size;
	double urand;
	double erand;
	double emean;
	struct timespec req, rem;
	struct timeval this, next;
	long long unsigned npackets_llu;
	int header_size = 0;

	to_write = snprintf(buf, sizeof(buf), "%s:u:%u:%u:%u:%llu+",
			    THRULAY_VERSION, client_port, opt.block_size,
			    protocol_rate, (long long unsigned)npackets);
	write_proposal(tcp_sock, buf, to_write);

	timer_start();

	response_size = read_response(tcp_sock, buf, sizeof(buf));
	if (strcmp(buf, "u:-") == 0)
		error(ERR_FATAL, "server rejected our UDP proposal");
	rc = sscanf(buf, "%u:%u:%u:%llu:%n", &server_port, &packet_size,
		    &protocol_rate, &npackets_llu, &n);
	if ((rc != 4) || (response_size != n+9) || (buf[n+8] != '+') ||
	    (packet_size > (int)sizeof(buf))) {
		error(ERR_FATAL, "malformed session response from server");
	}
	memcpy(nonce, buf+n, sizeof(nonce));
	npackets = (uint64_t)npackets_llu;

	/* Set target UDP server port and header size used for sending the
	 * UDP packets. */
	switch (server->sa_family) {
	case AF_INET:
		{
			struct sockaddr_in *sin = (struct sockaddr_in *)server;
			sin->sin_port = htons(server_port);

			header_size = IPV4_HEADER_SIZE + UDP_HEADER_SIZE;
		}
		break;
	case AF_INET6:
		{
			struct sockaddr_in6 *sin6 =
				(struct sockaddr_in6 *)server;
			sin6->sin6_port = htons(server_port);

			header_size = IPV6_HEADER_SIZE + UDP_HEADER_SIZE;
		}
		break;
	default:
		error(ERR_FATAL, "socket address family not supported.");
	}

	/* Disable keep-alives on the control TCP connection. */
	val = 0;
	rc = setsockopt(tcp_sock, SOL_SOCKET, SO_KEEPALIVE, (char *)&val,
	                sizeof(val));
	if (rc == -1)
		error(ERR_WARNING, "failed to disable keep-alives");

	/* Increase the buffer space of the sending UDP socket. */
	set_window_size_directed(udp_sock, opt.window, SO_SNDBUF);

	memset(buf, '2', sizeof(buf));	/* Nothing special about '2'. */
	memcpy(buf, nonce, sizeof(nonce));
	initstate(getpid() + getppid() + time(NULL),
		random_state, sizeof(random_state));

	/* RFC 2330: Chapter 11.1.3
	 * lambda = 1/emean */
	emean = 1000.0/(double)protocol_rate;

	/* Fill `next' timeval with current time */
	if (tsc_gettimeofday(&next) == -1) {
		perror("gettimeofday");
		error(ERR_FATAL, "gettimeofday(): failed");
	}
	normalize_tv(&next);

	for (packet = 0; packet < npackets; packet++) {
		/* RFC 2330: Chapter 11.1.3
		 * (Generating Poisson Sampling Intervals)
		 *
		 * Calculate Ui, a uniformly distributed (pseudo) random number
		 * between 0 and 1.
		 */
		urand = (double)((random()+1.0)/(RANDOM_MAX+1.0));
		assert(urand > 0 && urand <= 1);

		/* RFC 2330: Chapter 11.1.3
		 *
		 * Ei = -log(Ui) / lambda
		 */
		erand = -log(urand) * emean;

		/* Update `next' timeval. */
		next.tv_sec += floor(erand);
		next.tv_usec += (emean - floor(erand)) * 1000000;
		normalize_tv(&next);

		/* Calculate and set sequence number */
		msb = htonl(packet >> 32);	/* Most significant. */
		lsb = htonl(packet & ((1ULL<<32)-1));/* Least significant. */
		memcpy(buf+8, &msb, 4);
		memcpy(buf+12, &lsb, 4);

		if (opt.busywait) {
			/* Busy wait. */

			do {
				if (tsc_gettimeofday(&this) == -1) {
					perror("gettimeofday");
					error(ERR_FATAL, "gettimeofday(): "
							"failed");
				}
				normalize_tv(&this);
			} while (this.tv_sec < next.tv_sec ||
					(this.tv_sec == next.tv_sec &&
					 this.tv_usec < next.tv_usec));
		} else {
			/* No busy wait. */

			if (tsc_gettimeofday(&this) == -1) {
				perror("gettimeofday");
				error(ERR_FATAL, "gettimeofday(): failed");
			}
			normalize_tv(&this);

			/* Calculate how long to sleep. */
			req.tv_sec = next.tv_sec - this.tv_sec;
			req.tv_nsec = (next.tv_usec - this.tv_usec) * 1000;
			if (req.tv_nsec < 0) {
				--req.tv_sec;
				req.tv_nsec += 1000000000;
			}

			/* Only sleep if requested sleeping time is positive.
			 * If we have set a high rate with -u then we are
			 * maybe late with this packet. */
			if (req.tv_sec >= 0) {
				/* According to nanosleep(2), the nanosecond
				 * field has to be in the range 0 to 999999999.
				 */
				assert(req.tv_nsec >= 0);
				assert(req.tv_nsec <= 999999999);

				do {
					rc = nanosleep(&req, &rem);
					if (rc == -1) {
						if (errno != EINTR) {
							perror("nanosleep");
							error(ERR_FATAL,
								"nanosleep(): "
								"failed");
						}
						req.tv_sec = rem.tv_sec;
						req.tv_nsec = rem.tv_nsec;
					}
				} while (rc != 0);
			}
		}

		if (tsc_gettimeofday(&this) == -1) {
			perror("gettimeofday");
			error(ERR_FATAL, "gettimeofday(): failed");
		}
		normalize_tv(&this);

		tv2ntp(&this, buf+16);

		rc = sendto(udp_sock, buf, packet_size - header_size, 0,
		            server, server_len);
		if (rc == -1) {
			perror("sendto");
			error(ERR_FATAL, "sending UDP packet failed");
		}
	}

	timer_stop();

	sleep(1);	/* Let the UDP traffic drain. */
	close(udp_sock);

	printf("Client runtime: %8.3fs\n", timer.runtime);
}

void
thrulay_udp_exit (void)
{
	if (close(tcp_sock) == -1)
		error(ERR_WARNING, "thrulay_udp_exit(): Unable to close "
			"TCP connection socket.");

	/* Free memory of server sockaddr */
	free(server);
}

void
thrulay_udp_init (void)
{
	int rc;
	int tries;
	struct addrinfo hints, *res, *ressave;
	socklen_t addrlen;
	char service[7];

	tcp_sock = name2socket(opt.server_name, opt.port, 0, NULL,
			(void *)&server, &server_len);
	assert(tcp_sock > 0);
	read_greeting(tcp_sock);

	bzero(&hints, sizeof(struct addrinfo));
	hints.ai_flags = AI_PASSIVE;
	/* Use same address family as for TCP socket */
	hints.ai_family = server->sa_family;
	hints.ai_socktype = SOCK_DGRAM;

	client_port = UDP_PORT;
	tries = 0;
	do {
		client_port++;
		tries++;

		snprintf(service, sizeof(service), "%d", client_port);

		if ((rc = getaddrinfo(NULL, service, &hints,
						&res)) != 0) {
			fprintf(stderr, "getaddrinfo(): %s\n",
					gai_strerror(rc));
			exit(1);
		}
		ressave = res;

		do {
			udp_sock = socket(res->ai_family, res->ai_socktype,
					res->ai_protocol);
			if (udp_sock < 0)
				continue;

			/* Differentiated Services (DS) */
			if (opt.dscp) {
				rc = set_dscp(udp_sock, opt.dscp);
				if (rc == -1)
					error(ERR_WARNING,
							"thrulay_udp_init(): "
							"Unable to set DSCP "
							"value.");
			}

			if ((rc = bind(udp_sock, res->ai_addr,
							res->ai_addrlen)) == 0)
				break;		/* success */

			close(udp_sock);
		} while ((res = res->ai_next) != NULL);
	} while ((rc < 0) && (tries < TRY_UDP_PORTS));
	if (rc < 0)
		error(ERR_FATAL, "could not initialize UDP test");

	addrlen = res->ai_addrlen;

	freeaddrinfo(ressave);

	/* Protocol rate is in packets per 1000 seconds
	 *
	 * opt.rate is in bits per second:
	 *  => (opt.rate/8) = bytes/second
	 *  => (1000 * opt.rate/8) = bytes per 1000 seconds
	 */
	protocol_rate = ((1000/8) * opt.rate)/opt.block_size;

	/* npackets is number of packets to send in test
	 *
	 * npackets = protocol_rate / 1000 * opt.test_duration
	 */
	npackets = (opt.test_duration*opt.rate)/(8*opt.block_size);
}

void
thrulay_tcp_stop_id (int id)
{
	/* Delete stream socket in FD set. So it won't be processed in test
	 * loop. */
	FD_CLR(stream[id].sock, &rfds_orig);
	FD_CLR(stream[id].sock, &wfds_orig);

	/* Close our testing socket. This tells the server that this test
	 * has finished and the server will log test duration and average
	 * throughput. */
	thrulay_tcp_exit_id(id);
}

/* Stop TCP test. */
void
thrulay_tcp_stop (void)
{
	stop_test = 1;
}

/* Start TCP test. */
void
thrulay_tcp_start (void)
{
	int rc;
	int id;
	struct timeval tv, timeout;
	fd_set rfds, wfds;
	char buf[1024];

	timer_start();

	stop_test = 0;
	while (!stop_test) {
		rfds = rfds_orig;
		wfds = wfds_orig;

		timeout.tv_sec = 0;
		timeout.tv_usec = 10;

		rc = select(maxfd + 1, &rfds, &wfds, NULL, &timeout);
		if (rc < 0) {
			perror("select");
			error(ERR_FATAL, "select(): failed");
			/* NOTREACHED */
		}

		if (rc == 0) {
			/* Timeout occured. Check timer. */
			timer_check();
		}

		if (rc > 0) {
			/* Check all stream sockets. */
			for (id = 0; id < opt.num_streams; id++) {
				/* Read from socket. */
				if (FD_ISSET(stream[id].sock, &rfds)) {
					if (tsc_gettimeofday(&tv) == -1) {
						perror("gettimeofday");
						error(ERR_FATAL,
							"gettimeofday(): "
							"failed");
						/* NOTREACHED */
					}
	
					rc = read_exactly(stream[id].sock,
							buf, BLOCK_HEADER);
					assert(rc <= BLOCK_HEADER);
					if (rc == -1) {
						perror("read");
						error(ERR_WARNING, "premature "
								"end of test");
						thrulay_tcp_stop_id(id);
						break;
					}
					memcpy(&tv, buf, sizeof(tv));
					new_timestamp(id, &tv);
				}

				/* Write to socket */
				if (FD_ISSET(stream[id].sock, &wfds)) {
					if (tsc_gettimeofday(&tv) == -1) {
						perror("gettimeofday");
						error(ERR_FATAL,
							"gettimeofday(): "
							"failed");
						/* NOTREACHED */
					}
					memcpy(block, &tv, sizeof(tv));
					rc = write_exactly(stream[id].sock,
						block,
						(size_t)server_block_size);
					if (rc == -1) {
						perror("write_exactly");
						error(ERR_WARNING, "premature "
								"end of test");
						thrulay_tcp_stop_id(id);
					}
				} /* Write to socket */
			} /* for (id = 0; id < opt.num_streams; id++) */

			timer_check();
		} /* if (rc > 0) { */
	}

	timer_stop();
}

/* Deinitialize TCP stream with ID `id'. */
void
thrulay_tcp_exit_id (int id)
{
	/* Close connection to server. */
	if (close(stream[id].sock) == -1)
		error(ERR_WARNING, "thrulay_tcp_exit_id(): Unable to close "
				"server socket.");
}

/* Deinitialize all TCP streams. */
void
thrulay_tcp_exit (void)
{
	int id;

	for (id = 0; id < opt.num_streams; id++)
		thrulay_tcp_exit_id(id);

	/* Free memory of writing block. */
	free(block);
}

/* Some common MTU sizes with topology. */
struct _mtu_info {
	int mtu;
	char *top;
} mtu_list[] = {
	{ 65535,	"Hyperchannel" },		/* RFC1374 */
	{ 17914,	"16 MB/s Token Ring" },
	{ 16436,	"Linux Loopback device" },
	{ 8166,		"802.4 Token Bus" },		/* RFC1042 */
	{ 4464,		"4 MB/s Token Ring" },
	{ 4352,		"FDDI" },			/* RFC1390 */
	{ 1500,		"Ethernet/PPP" },		/* RFC894, RFC1548 */
	{ 1492,		"IEEE 802.3" },
	{ 1006,		"SLIP" },			/* RFC1055 */
	{ 576,		"X.25 & ISDN" },		/* RFC1356 */
	{ 296,		"PPP (low delay)" },
};
#define MTU_LIST_NUM	11

/* Calculate MTU out of MSS.
 * Set's global variable `mtu' and returns pointer to topology info.
 *
 * According to RFC879:
 * 	MSS = MTU - sizeof(TCPHDR) - sizeof(IPHDR)
 *
 * Where:
 * 	20 <= sizeof(IPHDR) <= 60
 *	20 <= sizeof(TCPHDR) <= 60
 *
 * This implies:
 *	MSS + 40 <= MTU <= MSS + 120
 */
char *
mtu_calc (int mss)
{
	int i;

#ifdef IP_MTU
	if (mtu) {
		for (i = 0; i < MTU_LIST_NUM; i++) {
			if (mtu == mtu_list[i].mtu) {
				return (mtu_list[i].top);
			}
		}
	}
	return "unknown";
#endif

	for (i = 0; i < MTU_LIST_NUM; i++) {
		if (((mss + 40) <= mtu_list[i].mtu)
				&& (mtu_list[i].mtu <= (mss + 120))) {
			mtu = mtu_list[i].mtu;
			return (mtu_list[i].top);
		}
	}

	/* No match. Return default one. */
	mtu = mss + 40;
	return "unknown";
}

/* Print test info before tests start. Displayed info includes local/remote
 * window, block size, MTU, MSS, test duration and reporting interval.
 */
void
thrulay_tcp_info (void)
{
	char *str_top = NULL;

	/* Print local/remote window and block size */
	printf("# local window = %dB; remote window = %dB\n",
			local_window, server_window);
	if (opt.block_size == server_block_size) {
		printf("# block size = %dB\n", opt.block_size);
	} else {
		printf("# requested block size = %dB; "
				"actual block size = %dB\n", opt.block_size,
				server_block_size);
	}

	str_top = mtu_calc(mss);

	/* Print MTU, MSS, topology info */
	printf("# MTU: %dB; MSS: %dB; Topology guess: %s\n", mtu, mss,
			str_top);
#ifdef IP_MTU
	printf("# MTU = getsockopt(IP_MTU); MSS = getsockopt(TCP_MAXSEG)\n");
#else
	if (!strcmp(str_top, "unknown")) {
		printf("# MTU = MSS + 40; MSS = getsockopt(TCP_MAXSEG)\n");
	} else {
		printf("# MTU = guessed out of MSS as in RFC 879; "
				"MSS = getsockopt(TCP_MAXSEG)\n");
	}
#endif

	/* Print test duration and reporting interval. */
        printf("# test duration = %ds; reporting interval = %ds\n",
			opt.test_duration, opt.reporting_interval);
	printf("#(ID) begin, s   end, s  Mb/s    RTT, ms: min  avg     max\n");

	fflush(stdout);
}

/* Initialize TCP stream with ID `id'. */
void
thrulay_tcp_init_id (int id)
{
	char buf[1024];
	int rc;
	int to_write;
	int my_server_block_size;
	int my_local_window, my_server_window;
	int my_mss;
	unsigned int len;
#ifdef IP_MTU
	int my_mtu;
#endif

	/* Open socket, do greeting */
	stream[id].sock = name2socket(opt.server_name, opt.port, opt.window,
			&my_local_window, NULL, NULL);
	assert(stream[id].sock > 0);
	read_greeting(stream[id].sock);
	to_write = snprintf(buf, sizeof(buf), "%s:t:%u:%u+", THRULAY_VERSION,
			opt.window, opt.block_size);
	assert(to_write > 0 && to_write < (int) sizeof(buf));
	write_proposal(stream[id].sock, buf, to_write);
	read_response(stream[id].sock, buf, sizeof(buf));
	my_server_window = my_server_block_size = -1;
        /* XXX: Very long numbers will not, of course, be processed
	 *      correctly by sscanf() below.  We could, if the number
	 *      exceeds a certain value (e.g., has more than a certain
	 *      number of characters, use some ``large'' -- still supported
	 *      -- value for window or block size.  It's not worth the
	 *      trouble. */
	rc = sscanf(buf, "%d:%d+", &my_server_window, &my_server_block_size);
	if (rc != 2)
		error(ERR_FATAL, "malformed session response from server");
	assert(my_server_window >= 0 && my_server_block_size >= 0);
	if (my_server_block_size < MIN_BLOCK)
		error(ERR_FATAL, "server gave block size less than MIN_BLOCK");
	if (my_server_block_size > MAX_BLOCK)
		error(ERR_FATAL, "server gave block size too large");
	if (my_server_window < 1500)
		error(ERR_WARNING, "server gave ridiculously small window");

#ifdef IP_MTU
	len = sizeof(my_mtu);
	if (getsockopt(stream[id].sock, SOL_IP, IP_MTU,
		(char *)&my_mtu, &len) == -1) {
		perror("getsockopt");
		error(ERR_WARNING, "unable to determine Path MTU");
	}
#endif

	/* Get Maximum Segment Size (MSS) */
	len = sizeof(my_mss);
	if (getsockopt(stream[id].sock, SOL_TCP, TCP_MAXSEG, (char *)&my_mss,
				&len) == -1) {
		perror("getsockopt");
		error(ERR_WARNING, "unable to determine TCP_MAXSEG");
	}

	/*
	 * Check/set local/remote window, server block size, MSS.
	 * As we display the local/remote window, server block size, MSS, MTU
	 * only once, we check that this information is the same for
	 * every stream.
	 */
	if (id == 0) {
		/* If this is first stream, initialize global values. */
		local_window = my_local_window;
		server_window = my_server_window;
		server_block_size = my_server_block_size;
		mss = my_mss;
#ifdef IP_MTU
		mtu = my_mtu;
#endif
	}
	if (local_window != my_local_window) {
		error(ERR_FATAL, "different local window");
	}
	if (server_window != my_server_window) {
		error(ERR_FATAL, "different server window");
	}
	if (server_block_size != my_server_block_size) {
		error(ERR_FATAL, "different server block size");
	}
	if (mss != my_mss) {
		error(ERR_FATAL, "different MSS");
	}
#ifdef IP_MTU
	if (mtu != my_mtu) {
		error(ERR_FATAL, "different MTU");
	}
#endif

	/* Differentiated Services (DS) */
	if (opt.dscp) {
		rc = set_dscp(stream[id].sock, opt.dscp);
		if (rc == -1)
			error(ERR_WARNING, "thrulay_tcp_init_id(): Unable to "
				"set DSCP value.");
	}

	assert((unsigned int)stream[id].sock < FD_SETSIZE);
}

/* Initialize all that has to do with TCP streams. */
void
thrulay_tcp_init (void)
{
	int id;

	tcp_stats_init();

	/* Clean read/write FD sets for select(). */
	FD_ZERO(&rfds_orig);
	FD_ZERO(&wfds_orig);

	for (id = 0; id < opt.num_streams; id++) {
		/* Initialize TCP stream. */
		thrulay_tcp_init_id(id);

		/* FD set */
		FD_SET(stream[id].sock, &rfds_orig);
		FD_SET(stream[id].sock, &wfds_orig);
		maxfd = (stream[id].sock > maxfd ? stream[id].sock : maxfd);
	}

	/* Allocate memory for writing block. */
	block = malloc((size_t)server_block_size);
	if (block == NULL) {
		error(ERR_FATAL, "malloc(): failed");
	}
	memset(block, '2', (size_t)server_block_size);

	thrulay_tcp_info();
}

/* Initialize thrulay. */
void
thrulay_init (void)
{
	tsc_init();

	/* If client has set no block size, set default ones. */
	if (!opt.block_size) {
		if (opt.rate) {
			/* Set default UDP block size */
			opt.block_size = 1500;
		} else {
			/* Set default TCP block size */
			opt.block_size = 8192;
		}
	}

	if (opt.rate) {
		thrulay_udp_init();
	} else {
		thrulay_tcp_init();
	}
}

/* Deinitialize thrulay. */
void
thrulay_exit (void)
{
	if (opt.rate) {
		thrulay_udp_exit();
	} else {
		thrulay_tcp_exit();
	}
}

void
thrulay_start (void)
{
	if (opt.rate) {
		thrulay_udp_start();
	} else {
		thrulay_tcp_start();
	}
}

void
thrulay_report_final (void)
{
	if (opt.rate) {
		thrulay_udp_report_final();
	} else {
		thrulay_tcp_report_final();
	}
}

/* 
 * Convert a string like "12k" to a number like 12000.  Return zero on
 * error.
 */
uint64_t
rate2i(char *s)
{
	uint64_t r;	/* Result. */
	char *p;
	int l;
	int suffix = 0;

	/* First, set the multiple. */
	l = strlen(s);
	switch (s[l-1]) {
	case '0':
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
	case '9':
		r = 1;
		break;
	case 'k':
	case 'K':
		r = 1000;
		suffix = 1;
		break;
	case 'm':
	case 'M':
		r = 1000000;
		suffix = 1;
		break;
	case 'g':
	case 'G':
		r = 1000000000ULL;
		suffix = 1;
		break;
	case 't':
	case 'T':
		r = 1000000000000ULL;
		suffix = 1;
		break;
	default:
		r = 0;
	}
	if (suffix)
		s[l-1] = '\0';
	r *= strtoll(s, &p, 10);
	if (!*s || *p)	/* Invalid character found. */
		r = 0;
	return r;
}

int
main(int argc, char *argv[])
{
	char *endptr = NULL;
	uint32_t tlng;
	int ch;

	options_init();

	while ((ch = getopt(argc, argv, "bD:m:t:i:w:l:p:u:")) != -1)
		switch (ch) {
		case 'b':
			opt.busywait = 1;
			break;
		case 'D':
			if (opt.dscp) {
				fprintf(stderr, "can only set \'-D\' once.\n");
				usage();
			}
			tlng = strtoul(optarg, &endptr, 0);
			/*
			 * Validate int conversion and verify user only sets
			 * last 6 bits (DSCP must fit in 6 bits - RFC 2474).
			 */
			if ((*endptr != '\0') || (tlng & ~0x3F)) {
				fprintf(stderr, "invalid value for option "
					"\'-D\'. DSCP value expected.\n");
				usage();
			}
			opt.dscp = (uint8_t)tlng;
			break;
		case 'm':
			opt.num_streams = atoi(optarg);
			if (opt.num_streams <= 0 ||
					opt.num_streams >= STREAMS_MAX) {
				fprintf(stderr, "number of test streams must "
						"be a positive integer < %d\n",
						STREAMS_MAX);
				usage();
			}
			break;
		case 't':
			opt.test_duration = atoi(optarg);
			if (opt.test_duration <= 0) {
				fprintf(stderr, "test duration must be "
					"a positive integer (in seconds)\n");
				usage();
			}
			break;
		case 'i':
			opt.reporting_interval = atoi(optarg);
			if (opt.reporting_interval <= 0) {
				fprintf(stderr, "reporting interval must be "
					"a positive integer (in seconds)\n");
				usage();
			}
			break;
		case 'w':
			opt.window = atoi(optarg);
			if (opt.window <= 0) {
				fprintf(stderr, "window must be "
					"a positive integer (in bytes)\n");
				usage();
			}
			break;
		case 'l':
			opt.block_size = atoi(optarg);
			if (opt.block_size <= 0) {
				fprintf(stderr, "block size must be "
					"a positive integer (in bytes)\n");
				usage();
			}
			break;
		case 'p':
			opt.port = atoi(optarg);
			if (opt.port <= 0) {
				fprintf(stderr, "port must be "
					"a positive integer\n");
				usage();
			}
			break;
		case 'u':
			opt.rate = rate2i(optarg);
			if (opt.rate == 0) {
				fprintf(stderr, "rate must be positive\n");
				usage();
			}
			break;
		default:
			usage();
		}
	argc -= optind;
	argv += optind;

	if (argc != 1)
		usage();
	opt.server_name = argv[0];
#if 0
	if (opt.rate && opt.reporting_interval) {
		fprintf(stderr, "No reporting interval with UDP tests.\n");
		usage();
	}
#endif

	/* Ignore SIGPIPE.  Always do in any socket code. */
	if (signal(SIGPIPE, SIG_IGN) == SIG_ERR) {
		perror("signal(SIGPIPE, SIG_IGN)");
		error(ERR_FATAL, "could not ignore SIGPIPE");
	}

	assert(sizeof(struct timeval) <= BLOCK_HEADER);

	thrulay_init();

	thrulay_start();

	thrulay_report_final();

	thrulay_exit();

	exit(0);
	/* NOTREACHED */
}
