/*
 * thrulayd.c -- network throughput tester (the server part).
 * 
 * Written by Stanislav Shalunov, http://www.internet2.edu/~shalunov/
 *            Bernhard Lutzmann, belu@users.sf.net
 * 
 * Copyright 2003, Internet2.
 * Legal conditions are in file LICENSE
 * (MD5 = ecfa50d1b0bfbb81b658c810d0476a52).
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <strings.h>
#include <signal.h>
#include <syslog.h>
#include <string.h>
#include <fcntl.h>
#include <math.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/param.h>
#include <sys/select.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <sys/wait.h>
#include <errno.h>
#include <time.h>
#include <sys/time.h>
#include <netdb.h>
#include "util.h"
#include "assertd.h"
#include "rcs.h"

RCS_ID("@(#) $Id: thrulayd.c,v 1.3 2005/10/17 23:22:18 shalunov Exp $")

#ifdef HAVE_FLOAT_H
#include <float.h>
#endif

#define min(a, b)	((a) < (b) ? (a) : (b))
#define max(a, b)	((a) > (b) ? (a) : (b))

/*
 * Reordering
 */
#define MAX_N		100
#define loop(x)		((x) >= 0 ? (x) : (x) + MAX_N)

/*
 * Access Control List (ACL)
 */
#define ACL_ALLOW	1
#define ACL_DENY	0

typedef struct acl {
	struct acl *next;
	struct sockaddr_storage sa;
	int mask;
} acl_t;

acl_t *acl_head = NULL;

/* ACL Prototypes */
int acl_allow_add (char *);
acl_t *acl_allow_add_list (acl_t *, struct sockaddr *, int);
int acl_check (struct sockaddr *);


/*
 * Logging
 */
#define LOGGING_MAXLEN	255		/* maximum string length */

enum {
	LOGTYPE_SYSLOG,
	LOGTYPE_STDERR
};

char timestr[20];
char *logstr = NULL;
int log_type = LOGTYPE_SYSLOG;		/* default is syslog */

/* Logging prototypes */
void logging_init (void);
void logging_exit (void);
void logging_log (int, const char *, ...);
void logging_log_string (int, const char *);
char *logging_time (void);

#define LISTEN_BACKLOG		128
#define INDICATOR		"thrulay"
#define THRULAY_VERSION		"2"
#define THRULAY_GREET		"thrulay/2+"
#define UDP_PORT		5003
#define TRY_UDP_PORTS		1000
#define UDP_BUFFER_SIZE		4194304
#define URANDOM_PATH		"/dev/urandom"

/*
 * Access Control List (ACL)
 */
/* Add address to list of allowed hosts. Returns 0 on success, -1 on failure. */
int
acl_allow_add (char *str)
{
	struct addrinfo hints, *res;
	char *pmask = NULL;
	int mask = -1;
	int rc;

	/* Extract netmask. */
	pmask = strchr(str, '/');
	if (pmask != NULL) {
		*pmask++ = '\0';
		mask = atoi(pmask);
	}

	bzero(&hints, sizeof(struct addrinfo));
	hints.ai_flags = AI_NUMERICHOST;
	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;

	if ((rc = getaddrinfo(str, NULL, &hints, &res)) != 0) {
		fprintf(stderr, "Error: getaddrinfo(): failed, %s\n",
				gai_strerror(rc));
		exit(1);
	}

	acl_head = acl_allow_add_list(acl_head, res->ai_addr, mask);

	freeaddrinfo(res);

	return 0;
}

acl_t *
acl_allow_add_list (acl_t *p, struct sockaddr *ss, int mask)
{
	if (p == NULL) {
		p = malloc(sizeof(acl_t));
		if (p == NULL) {
			perror("malloc");
			exit(1);
		}
		p->next = NULL;
		memcpy(&p->sa, ss, sizeof(struct sockaddr_storage));
		p->mask = mask;
	} else {
		p->next = acl_allow_add_list(p->next, ss, mask);
	}

	return p;
}

/* Checks if address in `sa' is allowed to connect. Returns ACL_ALLOW if
 * host is allowed to connect or ACL_DENY if host is not allowed to connect. */
int
acl_check (struct sockaddr *sa)
{
	struct sockaddr *acl_sa = NULL;
	struct sockaddr_in *sin = NULL, *acl_sin = NULL;
	struct sockaddr_in6 *sin6 = NULL, *acl_sin6 = NULL;
	acl_t *acl = NULL;
	int allow, i;

	/* If there are no hosts in ACL, then allow connection. */
	if (acl_head == NULL) {
		return ACL_ALLOW;
	}

	/* Check client sockaddr against entries in ACL */
	for (acl = acl_head; acl != NULL; acl = acl->next) {

		acl_sa = (struct sockaddr *)&acl->sa;

		/* Check that type matches. */
		if (sa->sa_family != acl_sa->sa_family) {
			continue;
		}

		switch (sa->sa_family) {
		case AF_INET:
			/* IPv4 address */
			sin = (struct sockaddr_in *)sa;
			acl_sin = (struct sockaddr_in *)acl_sa;

			/* Mask has been set to -1 if no netmask has been
			 * supplied. So we set default netmask here if
			 * that is the case. */
			if (acl->mask == -1) {
				acl->mask = 32;
			}

			/* Sanity check for mask. */
			if (acl->mask < 1 || acl->mask > 32) {
				fprintf(stderr, "Error: Bad netmask.\n");
				break;
			}

			if ((ntohl(sin->sin_addr.s_addr) >>
						(32 - acl->mask)) ==
					(ntohl(acl_sin->sin_addr.s_addr) >>
					 (32 - acl->mask))) {
				return ACL_ALLOW;
			}

			break;
		case AF_INET6:
			/* IPv6 address */
			sin6 = (struct sockaddr_in6 *)sa;
			acl_sin6 = (struct sockaddr_in6 *)acl_sa;

			/* Mask has been set to -1 if no netmask has been
			 * supplied. So we set default netmask here if
			 * that is the case. */
			if (acl->mask == -1) {
				acl->mask = 128;
			}

			/* Sanity check for netmask. */
			if (acl->mask < 1 || acl->mask > 128) {
				fprintf(stderr, "Error: Bad netmask.\n");
				break;
			}

			allow = 1;

			for (i = 0; i < (acl->mask / 8); i++) {
				if (sin6->sin6_addr.s6_addr[i]
					!= acl_sin6->sin6_addr.s6_addr[i]) {
					allow = 0;
					break;
				}
			}

			if ((sin6->sin6_addr.s6_addr[i] >>
			    (8 - (acl->mask % 8))) !=
					(acl_sin6->sin6_addr.s6_addr[i] >>
					 (8 - (acl->mask % 8)))) {
				allow = 0;
			}

			if (allow) {
				return ACL_ALLOW;
			}

			break;
		default:
			logging_log(LOG_WARNING, "Unknown address family.");
			break;
		}
	}

	return ACL_DENY;
}

/*
 * Logging
 */
/* Initialize logging. */
void
logging_init (void)
{
	/* Allocate memory for logging string. */
	logstr = malloc(LOGGING_MAXLEN);
	if (logstr == NULL) {
		fprintf(stderr, "Error: Unable to allocate memory for logging "
				"string.\n");
		exit(1);
	}

	switch (log_type) {
		case LOGTYPE_SYSLOG:
			openlog("thrulayd", LOG_NDELAY | LOG_CONS | LOG_PID,
					LOG_DAEMON);
			break;
		case LOGTYPE_STDERR:
			break;
	}
}

/* Deinitialize logging. */
void
logging_exit (void)
{
	switch (log_type) {
		case LOGTYPE_SYSLOG:
			closelog();
			break;
		case LOGTYPE_STDERR:
			break;
	}

	free(logstr);
}

/* Log message. Prepares input and hands it over to responsible log
 * facility. `priority' is the same priority as in syslog(3). */
void
logging_log (int priority, const char *fmt, ...)
{
	int n;
	va_list ap;

	memset(logstr, 0, LOGGING_MAXLEN);

	va_start(ap, fmt);
	n = vsnprintf(logstr, LOGGING_MAXLEN, fmt, ap);
	va_end(ap);

	/* If that worked, log the string. */
	if (n > -1 && n < LOGGING_MAXLEN)
		logging_log_string(priority, logstr);
}

void
logging_log_string (int priority, const char *s)
{
	switch (log_type) {
		case LOGTYPE_SYSLOG:
			syslog(priority, "%s", s);
			break;
		case LOGTYPE_STDERR:
			fprintf(stderr, "%s %s\n", logging_time(), s);
			fflush(stderr);
			break;
	}
}

/* Timestring will be stored in global variable `timestr'. Return pointer
 * to `timestr[0]'. */
char *
logging_time(void)
{
	time_t tp;
	struct tm *loc = NULL;

	/* Get current time. */
	tp = time(NULL);

	/* Convert it to local time representation. */
	loc = localtime(&tp);

	memset(&timestr, 0, sizeof(timestr));

	/* We use the format `month/day/year hrs:mins:secs'. Read strftime(3)
	 * if you want to change this. */
	strftime(&timestr[0], sizeof(timestr), "%m/%d/%Y %H:%M:%S", loc);

	return (&timestr[0]);
}

/*
 * Quantiles
 *
 * References:
 * [1] Manku, Rajagopalan, Lindsay: "Approximate Medians and other Quantiles
 *     in One Pass and with Limited Memory",
 *     http://www-db.stanford.edu/~manku/papers/98sigmod-quantiles.pdf
 */
#define QUANTILE_EPS	0.005

int quantile_b = 0;		/* number of buffers */
int quantile_k = 0;		/* number of elements in buffer */

double *quantile_buf = NULL;

double *quantile_input = NULL;	/* This is the buffer where the sequence of
				   incoming packets is saved. If we received
				   enough packets, we will write this buffer
				   to a quantile buffer. */
int quantile_input_cnt = 0;	/* number of elements in input buffer */
int quantile_empty_buffers = 0;	/* number of empty buffers */

int quantile_alternate = 0;	/* this is used to determine the offset in
				   COLLAPSE (if weight is even) */

uint64_t quantile_inf_cnt = 0;	/* this counter is for the additional -inf,
				 * +inf elements we added to NEW buffer to
				 * fill it up. */

typedef struct quantile {
	struct quantile *next;	/* pointer to next quantile buffer */
	int weight;		/* 0 if buffer is empty, > 0 if buffer is
				 * full */
	int level;
	double *buffer;
	int pos;		/* current position in buffer; used in
				   quantile_collapse() */
} quantile_t;

quantile_t *quantile_buffer_head = NULL;

/* Calculate binomial coefficient C(n, k). */
int64_t
binomial (int n, int k)
{
	int64_t bc = 0;
	int i, m;

	/* C(n, k) = C(n, n-k) */
	k = min(k, n-k);

	if (k >= 0) {
		bc = 1;
		m = max(k, n-k);

		for (i = 1; i <= k; i++) {
			bc = (bc * (m + i)) / i;
		}
	}

	return bc;
}

int
quantile_compare(const void *d1, const void *d2)
{
	if (*(double *)d1 < *(double *)d2) {
		return -1;
	} else if (*(double *)d1 == *(double *)d2) {
		return 0;
	} else {
		assert(*(double *)d1 > *(double *)d2);
		return 1;
	}
}

void
quantile_sort (double *list, int length)
{
	qsort(list, length, sizeof(double), quantile_compare);
}

/* Initialize quantiles. This sets values for `quantile_b' and `quantile_k'.
 * Then allocates memory for quantile buffers and input buffer.
 *
 * See section 4.5 "The New Algorithm" of paper [1] (see above). */
void
quantile_init (double eps, uint64_t N)
{
	quantile_t *qp = NULL;
	int b, b_tmp = 0;
	int k, k_tmp = 0;
	int h, h_max = 0;
	int i;

	/* "In practice, optimal values for b and k can be computed by trying
	 * out different values of b in the range 1 and 30." */
	for (b = 2; b <= 30; b++) {

		/* For each b, compute the largest integral h that satisfies:
		 *   (h-2) * C(b+h-2, h-1) - C(b+h-3, h-3) +
		 *                                 C(b+h-3, h-2) <= 2 * eps * N
		 */
		for (h = 0; ; h++) {
			if (((h-2) * binomial(b+h-2, h-1) -
						binomial(b+h-3, h-3) +
						binomial(b+h-3, h-2)) >
					(2 * eps * N)) {
				/* This h does not satisfy the inequality from
				 * above. */
				break;
			}
			h_max = h;
		}

		/* Now compute the smallest integral k that satisfies:
		 *   k * C(b+h-2, h-1) => N. */
		k = ceil(N / (double)binomial(b+h_max-2, h_max-1));

		/* Identify that b that minimizes b*k. */
		if ((b_tmp == 0) && (k_tmp == 0)) {
			/* Initialize values */
			b_tmp = b;
			k_tmp = k;
		}

		if ((b * k) < (b_tmp * k_tmp)) {
			/* Found b and k for which the product is smaller
			 * than for the ones before. Because we want to
			 * minimize b*k (required memory), we save them. */
			b_tmp = b;
			k_tmp = k;
		}
	}

	/* Set global quantile values. */
	quantile_b = b_tmp;
	quantile_k = k_tmp;

	/* Allocate memory for input buffer.
	 * We allocate enough space to save up to `2 * quantile_k' elements.
	 * This space is needed in the COLLAPSE policy if there are more
	 * than two empty buffers. Because then we have to invoke NEW on two
	 * buffers and thus need an input buffer with `2 * quantile_k'
	 * elements. */
	quantile_input = malloc(sizeof(double) * 2 * quantile_k);
	if (quantile_input == NULL) {
		logging_log(LOG_ERR, "malloc(): failed");
		exit(1);
	}
	quantile_input_cnt = 0;

	/* Allocate memory for output buffer.
	 * This buffer is used in COLLAPSE to store temporary output buffer
	 * before it gets copied to one of the buffers used in COLLAPSE. */
	quantile_buf = malloc(sizeof(double) * quantile_k);
	if (quantile_buf == NULL) {
		logging_log(LOG_ERR, "malloc(): failed");
		exit(1);
	}

	/* Allocate memory for quantile buffers. Buffers are linked lists
	 * with a pointer to next buffer.
	 * We need `quantile_b' buffers, where each buffer has space for
	 * `quantile_k' elements. */
	for (i = 0; i < quantile_b; i++) {
		if (i == 0) {
			/* Initialize first buffer. */
			qp = malloc(sizeof(quantile_t));
			if (qp == NULL) {
				logging_log(LOG_ERR, "malloc(): failed");
				exit(1);
			}
			quantile_buffer_head = qp;
		} else {
			qp->next = malloc(sizeof(quantile_t));
			if (qp->next == NULL) {
				logging_log(LOG_ERR, "malloc(): failed");
				exit(1);
			}
			qp = qp->next;
		}

		/* `qp' points to buffer that should be initialized. */
		qp->next = NULL;
		qp->weight = 0;		/* empty buffers have weight of 0 */
		qp->level = 0;
		qp->buffer = malloc(sizeof(double) * quantile_k);
		if (qp->buffer == NULL) {
			logging_log(LOG_ERR, "malloc(): failed");
			exit(1);
		}
	}

	/* Update number of empty quantile buffers. */
	quantile_empty_buffers = quantile_b;
}

/* Deinitialize quantiles.  Free's memory of quantile buffers and input
 * buffer. */
void
quantile_exit (void)
{
	quantile_t *qp = NULL, *next;

	qp = quantile_buffer_head;
	while (qp != NULL) {
		/* Save pointer to next buffer. */
		next = qp->next;

		/* Free buffer and list entry. */
		free(qp->buffer);
		free(qp);

		/* Set current buffer to next one. */
		qp = next;
	}

	quantile_buffer_head = NULL;

	/* Free output buffer. */
	free(quantile_buf);

	/* Free input buffer. */
	free(quantile_input);
}

/* Implementation of NEW operation from section 3.1 of paper [1].
 *
 * Takes as input an empty buffer. Simply populates the quantile buffer with
 * the next k elements from the input sequence, labels the buffer as full, and
 * assigns it a weight of 1.
 *
 * If there are not enough elements to fill up the buffer, we alternately add
 * -inf, +inf elements until buffer is full (-inf == 0, +inf == DBL_MAX).
 *
 * NOTE: We sort the elements in the input buffer before we copy them to
 * the quantile buffer. */
void
quantile_new (quantile_t *q, double *input, int k, int level)
{
	int i;

	/* Check that buffer is really empty. */
	if (q->weight != 0) {
		logging_log(LOG_ERR, "Error: NEW needs an empty buffer.");
		exit(1);
	}

	/* Sanity check for length of input sequence. */
	if (k > quantile_k) {
		logging_log(LOG_ERR, "Error: Bad input sequence length.");
		exit(1);
	}

	/* If there are not enough elements in the input buffer, fill it up
	 * with -inf, +inf elements. */
	for (i = k; i < quantile_k; i++) {
		if (i % 2) {
			input[i] = DBL_MAX;
		} else {
			input[i] = 0;
		}

		/* Increment counter that indicates how many additional
		 * elements we added to fill the buffer. */
		quantile_inf_cnt++;
	}

	quantile_sort(input, quantile_k);

	memcpy(q->buffer, input, sizeof(double) * quantile_k);

	/* Mark buffer as full and set level. */
	q->weight = 1;
	q->level = level;

	/* Update number of empty quantile buffers. */
	quantile_empty_buffers--;
}

/* Implementation of COLLAPSE operation from section 3.2 of paper [1].
 *
 * This is called from quantile_algorithm() if there are no empty buffers.
 * We COLLAPSE all the full buffers, where level has value `level'.
 * Output is written to the first buffer in linked list with level set to
 * `level'. The level of the output buffer is increased by 1.
 * All other buffers we used in the COLLAPSE are marked empty. */
void
quantile_collapse (int level)
{
	quantile_t *qp = NULL, *qp_out = NULL;
	int num_buffers = 0;	/* number of buffers with level `level' */
	int weight = 0;		/* weight of the output buffer */
	int offset;
	int i, j;
	double min_dbl;
	long next_pos;
	long merge_pos = 0;

	/* Check that there are at least two full buffers with given level.
	 * Also calculate weight of output buffer. */
	for (qp = quantile_buffer_head; qp != NULL; qp = qp->next) {
		if ((qp->weight != 0) && (qp->level == level)) {
			num_buffers++;
			weight += qp->weight;
			qp->pos = 0;
		} else {
			/* We mark the buffers that are not used in this
			 * COLLAPSE. */
			qp->pos = -1;
		}
	}
	if (num_buffers < 2) {
		logging_log(LOG_ERR, "Error: Not enough buffers for COLLAPSE.");
		exit(1);
	}

	/* NOTE: The elements in full buffers are sorted. So we don't have to
	 * do that again.
	 */
	/* Search for first full buffer with matching level. This is the buffer
	 * where we save the output. */
	for (qp_out = quantile_buffer_head; qp_out != NULL;
			qp_out = qp_out->next) {
		if (qp_out->pos != -1) {
			break;
		}
	}

	/* Calculate offset */
	if (weight % 2) {
		/* odd */
		offset = (weight + 1) / 2;
	} else {
		/* even - we alternate between two choices in each COLLAPSE */
		if (quantile_alternate % 2) {
			offset = weight / 2;
		} else {
			offset = (weight + 2)/ 2;
		}
		quantile_alternate = (quantile_alternate + 1) % 2;
	}

	/* Initialize next position of element to save. Because first position
	 * is at 0, we have to decrement offset by 1. */
	next_pos = offset - 1;

	for (i = 0; i < quantile_k; ) {

		/* Search for current minimal element in all buffers.
		 * Because buffers are all sorted, we just have to check the
		 * element at current position. */
		min_dbl = DBL_MAX;
		for (qp = quantile_buffer_head; qp != NULL; qp = qp->next) {
			/* Skip wrong buffers. */
			if (qp->pos == -1) {
				continue;
			}

			/* Check that we are not at the end of buffer. */
			if (qp->pos >= quantile_k) {
				continue;
			}

			/* Update minimum element. */
			min_dbl = min(min_dbl, qp->buffer[qp->pos]);
		}

		/* Now process this minimal element in all buffers. */
		for (qp = quantile_buffer_head; qp != NULL; qp = qp->next) {
			/* Skip wrong buffers. */
			if (qp->pos == -1) {
				continue;
			}

			/* Now process minimal element in this buffer. */
			for (; (qp->buffer[qp->pos] == min_dbl) &&
					(qp->pos < quantile_k); qp->pos++) {

				/* We run this loop `qp->weight' times.
				 * We check there if we are in a position
				 * so we have to save this element in our
				 * output buffer. */
				for (j = 0; j < qp->weight; j++) {

					if (next_pos == merge_pos) {
						quantile_buf[i] = min_dbl;
						i++;

						if (i == quantile_k) {
							/* We have written
							 * all elements to
							 * output buffer, so
							 * exit global loop. */
							goto out;
						}

						/* Update next position. */
						next_pos += weight;
					}

					merge_pos++;
				} /* for(j = 0; j < qp->weight; j++) */
			} /* for (; (qp->buffer[qp->pos] == min_dbl) &&
			     (qp->pos < quantile_k); qp->pos++) */
		} /* for (qp = quantile_buffer_head; qp!=NULL; qp = qp->next) */
	} /* for (i = 0; i < quantile_k; ) */

out:
	memcpy(qp_out->buffer, quantile_buf, sizeof(double) * quantile_k);

	/* Update weight of output buffer. */
	qp_out->weight = weight;
	qp_out->level = level+1;

	/* Update list of empty buffers. */
	for (qp = quantile_buffer_head; qp != NULL; qp = qp->next) {
		if ((qp->pos != -1) && (qp != qp_out)) {
			qp->weight = 0;
			qp->level = 0;
		}
	}
	quantile_empty_buffers += num_buffers - 1;
}

/* Implementation of OUTPUT operation from section 3.3 of paper [1].
 *
 * Takes c>= full input buffers in linked list. Returns phi-quantile.
 */
double
quantile_output (uint64_t npackets, double phi)
{
	quantile_t *qp = NULL;
	int num_buffers = 0;
	int weight = 0;
	int j;
	long next_pos = 0;
	long merge_pos = 0;
	double min_dbl;
	double beta;
	double phi2;		/* this is phi' */

	/* Check that there are at least two full buffers with given level. */
	for (qp = quantile_buffer_head; qp != NULL; qp = qp->next) {
		if (qp->weight != 0) {
			num_buffers++;
			weight += qp->weight;
			qp->pos = 0;
		} else {
			qp->pos = -1;
		}
	}
	if (num_buffers < 2) {
		/* XXX: In section 3.3 "OUTPUT operation" of paper [1] is says
		 * that OUTPUT takes c => 2 full input buffers. But what if we
		 * just have one full input buffer?
		 *
		 * For example this happens if you run a UDP test with a
		 * block size of 100k and a test duration of 3 seconds:
		 * $ ./thrulay -u 100k -t 3 localhost
		 */

		if (num_buffers != 1) {
			logging_log(LOG_ERR, "Error: Bad number of full "
					"buffers in OUTPUT operation.");
			exit(1);
		}
	}

	/* Calculate beta and phi' */
	beta = 1 + quantile_inf_cnt / (double)npackets;
	assert(beta >= 1.0);

	assert(phi >= 0.0 && phi <= 1.0);
	phi2 = (2 * phi + beta - 1) / (2 * beta);

	next_pos = ceil(phi2 * quantile_k * weight);

	/* XXX: If the client just sends a few packets, it is possible that
	 * next_pos is too large. If this is the case, decrease it. */
	if (next_pos >= (num_buffers * quantile_k)) {
		next_pos --;
	}

	while (1) {

		/* Search for current minimal element in all buffers.
		 * Because buffers are all sorted, we just have to check the
		 * element at current position. */
		min_dbl = DBL_MAX;
		for (qp = quantile_buffer_head; qp != NULL; qp = qp->next) {
			/* Skip wrong buffers. */
			if (qp->pos == -1) {
				continue;
			}

			/* Check that we are not at the end of buffer. */
			if (qp->pos >= quantile_k) {
				continue;
			}

			/* Update minimum element. */
			min_dbl = min(min_dbl, qp->buffer[qp->pos]);
		}

		/* Now process this minimal element in all buffers. */
		for (qp = quantile_buffer_head; qp != NULL; qp = qp->next) {
			/* Skip wrong buffers. */
			if (qp->pos == -1) {
				continue;
			}

			/* Now process minimal element in this buffer. */
			for (; (qp->buffer[qp->pos] == min_dbl) &&
					(qp->pos < quantile_k); qp->pos++) {

				/* Increment merge position `qp->weight'
				 * times. If we pass the position we seek,
				 * return current minimal element. */
				for (j = 0; j < qp->weight; j++) {
					if (next_pos == merge_pos) {
						return min_dbl;
					}
					merge_pos++;
				}
			}
		}
	}

	/* NOTREACHED */
}

/* Implementation of COLLAPSE policies from section 3.4 of paper [1].
 *
 * There are three different algorithms noted in the paper. We use the
 * "New Algorithm".
 */
void
quantile_algorithm (double *input, int k)
{
	quantile_t *qp = NULL, *qp2 = NULL;
	int min_level = -1;
	
	/* This should always be true. */
	if (quantile_buffer_head != NULL) {
		min_level = quantile_buffer_head->level;
	} else {
		logging_log(LOG_ERR, "Error: quantiles not initialized.");
		exit(1);
	}

	/* Get minimum level of all currently full buffers. */
	for (qp = quantile_buffer_head; qp != NULL; qp = qp->next) {
		if (qp->weight != 0) {
			/* Full buffer. */
			min_level = min(min_level, qp->level);
		}
	}

	if (quantile_empty_buffers == 0) {
		/* There are no empty buffers. Invoke COLLAPSE on the set
		 * of buffers with minimum level. */

		quantile_collapse(min_level);
	} else if (quantile_empty_buffers == 1) {
		/* We have exactly one empty buffer. Invoke NEW and assign
		 * it level `min_level'. */

		/* Search the empty buffer. */
		for (qp = quantile_buffer_head; qp != NULL; qp = qp->next) {
			if (qp->weight == 0) {
				/* Found empty buffer. */
				break;
			}
		}

		quantile_new(qp, input, k, min_level);
	} else {
		/* There are at least two empty buffers. Invoke NEW on each
		 * and assign level `0' to each. */

		/* Search for two empty buffers. */
		for (qp = quantile_buffer_head; qp != NULL; qp = qp->next) {
			if (qp->weight == 0) {
				/* Found first empty buffer. */
				break;
			}
		}
		for (qp2 = qp->next; qp2 != NULL; qp2 = qp2->next) {
			if (qp2->weight == 0) {
				/* Found second empty buffer. */
				break;
			}
		}

		if (k <= quantile_k) {
			/* This could happen if we call this after we
			 * received all packets but don't have enough to
			 * fill up two buffers. */

			quantile_new(qp, input, k, 0);
		} else {
			/* We have enough input data for two buffers. */
			quantile_new(qp, input, quantile_k, 0);
			quantile_new(qp2, input+quantile_k, k - quantile_k, 0);
		}
	}
}

void __attribute__((noreturn))
usage(void)
{
	fprintf(stderr, "Usage: thrulayd [-a address ] [-w#] [-p#] [-d]\n");
	fprintf(stderr, "\t-a address\tadd address to list of allowed hosts "
			"(CIDR syntax)\n");
	fprintf(stderr, "\t-w#\t\twindow, in bytes (default: 4194304B)\n");
	fprintf(stderr, "\t-p#\t\tserver port (default: 5003)\n");
	fprintf(stderr, "\t-d \t\tdebug (no daemon, log to stderr)\n");
	exit(1);
}

void
sighandler(int sig)
{
	int status;

	switch (sig) {
	case SIGCHLD:
		while (waitpid(-1, &status, WNOHANG) > 0)
			;
		break;
	case SIGHUP:
		logging_log(LOG_NOTICE, "got SIGHUP, don't know what do do");
		break;
	case SIGPIPE:
		break;
	default:
		logging_log(LOG_ALERT, "got signal %d, but don't remember "
				"intercepting it, aborting", sig);
		abort();
	}
}

char *
sock_ntop (const struct sockaddr *sa)
{
	static char str[128];
	struct sockaddr_in *sin = NULL;
	struct sockaddr_in6 *sin6 = NULL;

	switch (sa->sa_family) {
	case AF_INET:
		sin = (struct sockaddr_in *)sa;

		if (inet_ntop(AF_INET, &sin->sin_addr, str,
					sizeof(str)) == NULL)
			return NULL;

		return (str);
	case AF_INET6:
		sin6 = (struct sockaddr_in6 *)sa;

		if (inet_ntop(AF_INET6, &sin6->sin6_addr, str,
					sizeof(str)) == NULL)
			return NULL;

		return (str);
	default:
		return NULL;
	}
}

void
log_client_address (const struct sockaddr *sa)
{
	logging_log(LOG_NOTICE, "connection from %s", sock_ntop(sa));
}

void
udp_test(int fd, char* proposal)
{
	int rc;
	int opt;
	unsigned int client_port;
	unsigned int packet_size;
	int udp_buffer_size;
	uint64_t rate;			/* In packets per 1000 seconds. */
	uint64_t npackets;		/* Proposed number of packets. */
	uint64_t packets_received = 0;	/* Unique packets received. */
	uint64_t packets_sent = 0;	/* What the client said at the end. */
	uint64_t packets_total = 0;	/* Total packets received, including
					   duplicates. */
	char nonce[8];
	char buffer[65536];
	int response_size;
	int sock;
	/* Decrement will be canceled before first use. */
	int server_port = UDP_PORT - 1;
	struct sockaddr *cliaddr = NULL;
	socklen_t addrlen, len;
	struct addrinfo hints, *res, *ressave;
	char service[7];
	int tries;
	int random_fd;
	int nfds;
	fd_set readfds;
	int test_done = 0;
	int graceful_termination = 0;
	double delay;
	struct timeval tv, tvc;
	long long unsigned rate_llu, npackets_llu, packets_sent_llu;
	uint64_t packet_sqn;		/* Packet sequence number */
	uint32_t msb, lsb;		/* 32bit parts of sequence number */
	/* reordering */
	uint64_t m[MAX_N];		/* We have m[j-1] == number of
					 * j-reordered packets. */
	uint64_t ring[MAX_N];		/* Last sequence numbers seen */
	int r = 0;			/* Ring pointer for next write. */
	int64_t l = 0;			/* Counter of sequence numbers. */
	int j;
	char buffer_reord[MAX_N * 80];
	/* duplication */
	uint64_t packets_dup = 0;	/* Duplicated packets */
	uint64_t *bitfield = NULL;	/* Bit field used to check for
					   duplicated packets. */
	uint64_t bitfield_len = 0;	/* Number of sectors in bitfield */
	uint64_t bitfield_sec;		/* Which sector in bitfield */
	uint64_t bitfield_bit;		/* Which bit in sector */
	/* quantiles */
	double quantile_00;
	double quantile_50;
	double quantile_95;

	rc = sscanf(proposal, "%u:%u:%llu:%llu+", &client_port, &packet_size,
		    &rate_llu, &npackets_llu);
	if (rc != 4) {
		logging_log(LOG_WARNING, "malformed parameters in UDP "
			    "proposal");
		goto log;
	}
	rate = (uint64_t)rate_llu;
	npackets = (uint64_t)npackets_llu;

	bzero(&hints, sizeof(struct addrinfo));
	hints.ai_flags = AI_PASSIVE;
	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = SOCK_DGRAM;

	/* Get a port number for the server. */
	tries = 0;
	do {
		server_port++;
		tries++;

		/* Convert integer port number to string for getaddrinfo(). */
		snprintf(service, sizeof(service), "%d", server_port);

		if ((rc = getaddrinfo(NULL, service, &hints, &res)) != 0) {
			logging_log(LOG_ERR, "getaddrinfo(): failed, %s",
					gai_strerror(rc));
			exit(1);
		}
		ressave = res;

		do {
			sock = socket(res->ai_family, res->ai_socktype,
					res->ai_protocol);
			if (sock < 0)
				continue;

			if ((rc = bind(sock, res->ai_addr,
							res->ai_addrlen)) == 0)
				break;		/* success */

			close(sock);
		} while ((res = res->ai_next) != NULL);
	} while ((rc < 0) && (tries < TRY_UDP_PORTS));
	if (rc < 0) {
		logging_log(LOG_WARNING, "could not bind a UDP socket");
		goto log;
	}

	addrlen = res->ai_addrlen;

	freeaddrinfo(ressave);

	cliaddr = malloc(addrlen);
	if (cliaddr == NULL) {
		logging_log(LOG_ERR, "malloc(): failed");
		exit(1);
	}

	/* Obtain random data for the nonce. */
	random_fd = open(URANDOM_PATH, O_RDONLY);
	if (random_fd < 0) {
		logging_log(LOG_WARNING, "could not open random device");
		goto log;
	}
	rc = read(random_fd, nonce, sizeof(nonce));
	if (rc != sizeof(nonce)) {
		logging_log(LOG_WARNING, "could not read urandom for UDP "
				"nonce");
		goto log;
	}
	close(random_fd);
	/* 
	 * Build a UDP session response.  Note that since nonce may
	 * contain NUL characters, it is not a C string.  The length
	 * is kept in response_size.
	 */
	sprintf(buffer, "%u:%u:%llu:%llu:", server_port, packet_size,
		(long long unsigned)rate, (long long unsigned)npackets);
	response_size = strlen(buffer);
	memcpy(buffer + response_size, nonce, sizeof(nonce));
	response_size += sizeof(nonce);
	buffer[response_size] = '+';
	response_size++;

	/* Disable keep-alives on the control TCP connection. */
	opt = 0;
	rc = setsockopt(fd, SOL_SOCKET, SO_KEEPALIVE, (char *)&opt,
			sizeof(opt));
	if (rc == -1) {
		logging_log(LOG_WARNING, "setsockopt(SO_KEEPALIVE): failed, "
				"continuing");
	}

	/* Increase the buffer space of the receiving UDP socket. */
	udp_buffer_size = set_window_size_directed(sock, UDP_BUFFER_SIZE,
						   SO_RCVBUF);

	write_exactly(fd, buffer, response_size);

	/* Set sockets to nonblocking mode: will use select(). */
	opt = fcntl(fd, F_GETFL);
	if (opt == -1) {
		logging_log(LOG_WARNING, "fcntl(F_GETFL): failed, continuing");
	} else {
		/* Set TCP socket to nonblocking mode. */
		rc = fcntl(fd, F_SETFL, opt | O_NONBLOCK);
		if (rc == -1) {
			logging_log(LOG_WARNING, "fcntl(F_SETFL): failed, "
					"continuing");
		}
	}

	opt = fcntl(sock, F_GETFL);
	if (opt == -1) {
		logging_log(LOG_WARNING, "fcntl(F_GETFL): failed, continuing");
	} else {
		/* Set UDP socket to nonblocking mode. */
		rc = fcntl(sock, F_SETFL, opt | O_NONBLOCK);
		if (rc == -1) {
			logging_log(LOG_WARNING, "fcntl(F_SETFL): failed, "
					"continuing");
		}
	}

	/* Initialize reordering */
	for (j = 0; j < MAX_N; j++)
		m[j] = 0;

	/* Initialize quantiles */
	quantile_init(QUANTILE_EPS, npackets);

	/* Allocate memory for bit field */
	bitfield_len = ((npackets % 64) ? (npackets / 64 + 1) : npackets / 64);
	bitfield = calloc(bitfield_len, sizeof(uint64_t));
	if (bitfield == NULL) {
		perror("calloc");
		error(ERR_FATAL, "calloc(): failed");
		/* NOTREACHED */
	}

	/*
	 * Could have set nfds to FD_SETSIZE, but this is marginally
	 * more efficient.
	 */
	nfds = (fd > sock? fd: sock) + 1;
	while (! test_done) {
		FD_ZERO(&readfds);
		FD_SET(fd, &readfds);
		FD_SET(sock, &readfds);
		rc = select(nfds, &readfds, 0, 0, 0);
		if (FD_ISSET(sock, &readfds)) {
			/* A new UDP packet arrived. */
			/* XXX: Need to loop in case of coalescence. */
			len = addrlen;
			rc = recvfrom(sock, buffer, sizeof(buffer), 0,
					cliaddr, &len);
			if (rc < 0) {
				if (errno != EAGAIN) {
					logging_log(LOG_WARNING, "recvfrom(): "
							"failed");
				}
			}
			if (memcmp(buffer, nonce, sizeof(nonce)) == 0) {
				/* Legitimate packet -- nonce matches.
				 * This includes duplicated packets. */
				packets_total++;

				/*
				 * Calculate packet sequence number
				 */
				memcpy(&msb, buffer+8, 4);
				memcpy(&lsb, buffer+12, 4);

				msb = ntohl(msb);
				lsb = ntohl(lsb);

				packet_sqn = (uint64_t)msb << 32;
				packet_sqn |= lsb;

				/* Sanity check for packet sequence number */
				if (packet_sqn > (npackets - 1)) {
					error(ERR_WARNING, "Packet has bad "
							"sequence number, "
							"skipping.");
					continue;
				}

				/*
				 * Duplication
				 */
				/* Calculate sector */
				bitfield_sec = packet_sqn >> 6;

				/* Calculate bit in sector */
				bitfield_bit = (uint64_t)1 << (packet_sqn & 63);

				if (bitfield[bitfield_sec] & bitfield_bit) {
					/* Duplicated packet */
					packets_dup++;
					continue;
				} else {
					/* Unique packet */
					bitfield[bitfield_sec] |= bitfield_bit;
					packets_received++;
				}

				/*
				 * Calculate delay for this packet
				 */
				/* Get current time for packet delay */
				if (tsc_gettimeofday(&tv) == -1) {
					perror("gettimeofday");
					error(ERR_FATAL, "gettimeofday(): "
							"failed");
					/* NOTREACHED */
				}
				normalize_tv(&tv);

				/* Get client gettimeofday() for packet */
				ntp2tv(&tvc, buffer+16);

				delay = time_diff(&tvc, &tv);

				/*
				 * Quantiles
				 */
				quantile_input[quantile_input_cnt++] = delay;

				/* If we have at least two empty buffers,
				 * we need input for two buffers, to twice
				 * the value of `quantile_k'. */
				if (quantile_empty_buffers >= 2) {
					if (quantile_input_cnt ==
							(2 * quantile_k)) {
						quantile_algorithm(
							quantile_input,
							quantile_input_cnt);
						/* Reset counter. */
						quantile_input_cnt = 0;
					}
				} else {
					/* There are 0 or 1 empty buffers */
					if (quantile_input_cnt == quantile_k) {
						quantile_algorithm(
							quantile_input,
							quantile_input_cnt);
						/* Reset counter. */
						quantile_input_cnt = 0;
					}
				}

				/*
				 * Reordering
				 */
				for (j = 0; j < min(l, MAX_N) &&
						packet_sqn < ring[loop(r-j-1)];
						j++) {
					m[j]++;
				}
				ring[r] = packet_sqn;
				l++;
				r = (r + 1) % MAX_N;
			} else {
				logging_log(LOG_WARNING, "nonce does not "
						"match");
			}
		}
		if (FD_ISSET(fd, &readfds)) {
			/* Data on the TCP connection -- test done. */
			test_done = 1;
			/* XXX: Assume it comes through in one piece. */
			rc = read(fd, buffer, sizeof(buffer));
			if (rc < 0) {
				if (errno != EAGAIN) {
					logging_log(LOG_WARNING, "could not "
							"read TCP connection "
							"at the end of a UDP "
							"test");
					goto log;
				}
			}
			if (buffer[0] == '+') {
				rc = sscanf(buffer+1, "%llu:",
					    &packets_sent_llu);
				packets_sent = (uint64_t)packets_sent_llu;
				if (rc != 1) {
					logging_log(LOG_WARNING, "malformed "
							"end of test");
					goto log;
				}
				graceful_termination = 1;
			}
		}
	}
	/* 
	 * Done with the UDP test.  Set the TCP socket back to
	 * blocking mode.
	 */
	opt = fcntl(fd, F_GETFL);
	if (opt == -1) {
		logging_log(LOG_WARNING, "fcntl(F_GETFL): failed, continuing");
	} else {
		/* Set TCP socket back to blocking mode */
		rc = fcntl(fd, F_SETFL, opt & ~O_NONBLOCK);
		if (rc == -1) {
			logging_log(LOG_WARNING, "fcntl(F_SETFL): failed, "
					"continuing");
		}
	}

	/*
	 * Quantiles
	 *
	 * Perform last algorithm operation with a possibly not full
	 * input buffer.
	 */
	if (quantile_input_cnt > 0) {
		quantile_algorithm(quantile_input, quantile_input_cnt);
	}

	quantile_00 = quantile_output(packets_received, 0.00);
	quantile_50 = quantile_output(packets_received, 0.50);
	quantile_95 = quantile_output(packets_received, 0.95);

	/*
	 * Write reordering statistics to buffer
	 */
	bzero(&buffer_reord, sizeof(buffer_reord));
	r = 0;
	for (j = 0; j < MAX_N && m[j]; j++) {
		r += snprintf(buffer_reord + r, sizeof(buffer_reord) - r,
				"  %d-reordering = %3.3f%%\n", j + 1,
				100.0 * m[j] / (l - j - 1));
	}

	if (j == 0) {
		snprintf(buffer_reord + r, sizeof(buffer_reord) - r,
				"  no reordering\n");
	} else if (j < MAX_N) {
		snprintf(buffer_reord + r, sizeof(buffer_reord) - r,
				"  no %d-reordering\n", j + 1);
	} else {
		snprintf(buffer_reord + r, sizeof(buffer_reord) - r,
				"  only up to %d-reordering is "
				"handled\n", MAX_N);
	}

	snprintf(buffer, sizeof(buffer),
		 "server used UDP buffer size of %d bytes\n"
		 "client proposed sending %llu packets\n"
		 "client said it sent %llu packets\n"
		 "server received %llu total packets\n"
		 "server received %llu unique packets\n"
		 "duplication: %llu packets\n"
		 "quantiles (ignoring clock offset):\n"
		 "   0th: %8.3fms\n"
		 "  50th: %8.3fms\n"
		 "  95th: %8.3fms\n"
		 "packet loss: %3.3f%%\n"
		 "reordering:\n"
		 "%s",
		 udp_buffer_size,
		 (long long unsigned)npackets,
		 (long long unsigned)packets_sent,
		 (long long unsigned)packets_total,
		 (long long unsigned)packets_received,
		 (long long unsigned)packets_dup,
		 1000.0 * quantile_00,
		 1000.0 * quantile_50,
		 1000.0 * quantile_95,
		 packets_sent > packets_received?
		 (100.0*(packets_sent - packets_received))/packets_sent: 0,
		 buffer_reord);

	write_exactly(fd, buffer, strlen(buffer));

	quantile_exit();
 log:
	free(bitfield);
	free(cliaddr);
	return;
}

void
serve_client(int fd)
{
	int rc;
	char buffer[1024];
	char *buf;
	char *block = NULL;
	int w, b;
	int to_write;
	int blocks = 0;
	struct timeval start = {0, 0};

	buf = buffer;
	rc = write_exactly(fd, THRULAY_GREET, sizeof THRULAY_GREET - 1);
	assert(rc < (int) sizeof THRULAY_GREET);
	if (rc < (int) sizeof THRULAY_GREET - 1) {
		if (rc == -1) {
			logging_log(LOG_WARNING, "write(greeting): failed");
		}
		logging_log(LOG_WARNING, "could not write greeting, exiting");
		goto log;
	}
	/* XXX: Assume that few-byte session proposal will come in one
           TCP packet and read in one block. */
	rc = read(fd, buf, sizeof(buffer) - 1);
	assert(rc < (int) sizeof(buffer));
	if (rc == -1) {
		logging_log(LOG_NOTICE, "could not read session proposal");
		goto log;
	}
	assert(rc > 0);
	buf[rc] = '\0';
	w = b = -1;
	rc = memcmp(buf, INDICATOR, sizeof(INDICATOR) - 1);
	if (rc != 0) {
		logging_log(LOG_WARNING, "malformed protocol indicator");
		goto log;
	}
	buf += sizeof(INDICATOR) - 1;	/* Skip the indicator. */
	if (*buf != '/') {
		logging_log(LOG_WARNING, "protocol indicator not followed by "
				"'/'");
		goto log;
	}
	buf++;				/* Skip the slash. */
	rc = memcmp(buf, THRULAY_VERSION, sizeof(THRULAY_VERSION) - 1);
	if (rc != 0) {
		logging_log(LOG_WARNING, "malformed protocol version");
		goto log;
	}
	buf += sizeof(THRULAY_VERSION) - 1;	/* Skip the version. */
	if (*buf != ':') {
		logging_log(LOG_WARNING, "protocol version not followed by "
				"':'");
		goto log;
	}
	buf++;				/* Skip the colon. */
	if ((buf[0] == 'u') && (buf[1] == ':')) {
		buf += 2;		/* Skip the test type. */
		udp_test(fd, buf);
		return;
	} else if ((buf[0] != 't') || (buf[1] != ':')) {
		logging_log(LOG_WARNING, "unknown test proposal type");
		goto log;
	}
	buf += 2;			/* Skip "t:". */
	/* Do TCP test. */
	/* XXX: Very long numbers will not, of course, be processed
           correctly by sscanf() below.  We could, if the number
           exceeds a certain value (e.g., has more than a certain
           number of characters, use some ``large'' -- still supported
           -- value for window or block size.  It's not worth the
           trouble. */
	rc = sscanf(buf, "%d:%d+", &w, &b);
	if (rc != 2) {
		logging_log(LOG_WARNING, "malformed session proposal from "
				"client");
		goto log;
	}
	assert(w >= 0 && b >= 0);	/* Unsigned formats. */
	if (b < MIN_BLOCK)
		b = MIN_BLOCK;
	if (b > MAX_BLOCK)
		b = MAX_BLOCK;
	if (w < 1500)
		w = 1500;
	block = malloc((size_t) b);
	if (! block) {
		logging_log(LOG_ALERT, "could not allocate memory");
		goto log;
	}
	w = set_window_size(fd, w);
	to_write = snprintf(buffer, sizeof(buffer), "%u:%u+", w, b);
	assert(to_write > 0 && to_write < (int) sizeof(buffer));
	rc = write_exactly(fd, buffer, (size_t) to_write);
	assert(rc <= to_write);
	tsc_gettimeofday(&start);

	/* And now the testing loop. */
	while ((rc = read_exactly(fd, block, (size_t) b)) == b) {
		blocks++;
		rc = write_exactly(fd, block, BLOCK_HEADER);
		if (rc != BLOCK_HEADER) {
			if (rc == -1) {
				if (errno == ECONNRESET) {
					/* It's possible that this is the last
					 * packet the client sent in test.
					 * So the client has maybe closed the
					 * TCP socket and we are unable to send
					 * back this header and get an error
					 * message here.
					 * If this is the case, don't print an
					 * error message. */
					goto log;
				}
				logging_log(LOG_NOTICE, "write(block_header): "
						"failed");
			}
			logging_log(LOG_NOTICE, "could not write block header");
			goto log;
		}
	}
	if (rc == -1) {
		if (errno != ECONNRESET) {
			logging_log(LOG_NOTICE, "while testing: read(): "
					"failed");
		}
	} else {
		assert(rc == 0);
	}
 log:
	{
		struct timeval end;
		double bytes, seconds;

		if (start.tv_sec == 0 && start.tv_usec == 0) {
			logging_log(LOG_NOTICE, "nothing transfered");
			return;
		}
		tsc_gettimeofday(&end);
		seconds = time_diff(&start, &end);
		bytes = (double) b * (double) blocks;
		logging_log(LOG_NOTICE, "test duration = %.3fs, "
				"average throughput = %.3fMb/s",
				seconds, bytes*8.0/seconds/1000000.0);
	}
	free(block);

	if (close(fd) == -1)
		logging_log(LOG_WARNING, "close(): failed");
}

int
main(int argc, char *argv[])
{
	int window = 4194304;		/* Window size, in bytes. */
	int port = 5003;		/* Server TCP port number. */
	char service[7];
	int on = 1;
	int listenfd, n;
	struct addrinfo hints, *res, *ressave;
	socklen_t addrlen, len;
	struct sockaddr *cliaddr = NULL;
	int ch;
	int argcorig = argc;

	while ((ch = getopt(argc, argv, "a:w:p:d")) != -1) {
		switch (ch) {
		case 'a':
			if (acl_allow_add(optarg) == -1) {
				fprintf(stderr, "unable to add host to ACL "
						"list\n");
				usage();
			}
			break;
		case 'w':
			window = atoi(optarg);
			if (window <= 0) {
				fprintf(stderr, "window must be "
					"a positive integer (in bytes)\n");
				usage();
			}
			break;
		case 'p':
			port = atoi(optarg);
			if (port <= 0) {
				fprintf(stderr, "port must be "
					"a positive integer\n");
				usage();
			}
			break;
		case 'd':
			/* Activate debugging mode. */
			log_type = LOGTYPE_STDERR;
			break;
		default:
			usage();
		}
	}
	argc = argcorig;

	argc -= optind;
	argv += optind;

	if (argc != 0)
		usage();

	/* Ignore SIGPIPE.  Always do in any socket code. */
	if (signal(SIGPIPE, SIG_IGN) == SIG_ERR) {
		perror("ignoring SIGPIPE");
		error(ERR_FATAL, "could not ignore SIGPIPE");
		/* NOTREACHED */
	}
	{
		struct sigaction sa;

		sa.sa_handler = sighandler;
		sa.sa_flags = 0;
		sigemptyset (&sa.sa_mask);
		sigaction (SIGHUP, &sa, NULL);
		sigaction (SIGALRM, &sa, NULL);
		sigaction (SIGCHLD, &sa, NULL);
	}

	logging_init();

	tsc_init();

	bzero(&hints, sizeof(struct addrinfo));
	hints.ai_flags = AI_PASSIVE;
	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;

	/* Convert integer port number to string for getaddrinfo(). */
	snprintf(service, sizeof(service), "%d", port);

	if ((n = getaddrinfo(NULL, service, &hints, &res)) != 0) {
		fprintf(stderr, "Error: getaddrinfo() failed: %s\n",
				gai_strerror(n));
		exit(1);
	}
	ressave = res;

	do {
		listenfd = socket(res->ai_family, res->ai_socktype,
				res->ai_protocol);
		if (listenfd < 0)
			continue;

		if (setsockopt(listenfd, SOL_SOCKET, SO_REUSEADDR,
					(char *)&on, sizeof(on)) == -1) {
			perror("setsockopt");
			error(ERR_WARNING, "setsockopt(SO_REUSEADDR): failed, "
					"continuing");
		}

		if (setsockopt(listenfd, SOL_SOCKET, SO_KEEPALIVE,
				(char *)&on, sizeof(on)) == -1) {
			perror("setsockopt");
			error(ERR_WARNING, "setsockopt(SO_KEEPALIVE): failed, "
					"continuing");
		}

		window = set_window_size(listenfd, window);

		if (bind(listenfd, res->ai_addr, res->ai_addrlen) == 0)
			break;		/* success */

		close(listenfd);	/* bind error, close and try next one */
	} while ((res = res->ai_next) != NULL);

	if (res == NULL) {
		error(ERR_FATAL, "unable to start server");
		/* NOTREACHED */
	}

	if (listen(listenfd, LISTEN_BACKLOG) < 0) {
		perror("listen");
		exit(1);
	}

	/* Save size of client address */
	addrlen = res->ai_addrlen;

	freeaddrinfo(ressave);
	
	/* Allocate memory for client address */
	cliaddr = malloc(addrlen);
	if (cliaddr == NULL) {
		perror("malloc");
		exit(1);
	}

	/* Daemonize thrulayd */
	if (log_type == LOGTYPE_SYSLOG) {
		if (daemon(0, 0) == -1) {
			perror("daemon");	/* It could be going to
						   /dev/null... */
			exit(1);
			/* NOTREACHED */
		}
	}

	logging_log(LOG_NOTICE, "thrulayd started, listening on port %d with "
			"window size %d", port, window);

	/* Main loop. */
	while (1) {
		int fd, pid;

		len = addrlen;

		fd = accept(listenfd, cliaddr, &len);
		if (fd == -1) {
			if (errno != EINTR) {
				logging_log(LOG_WARNING, "accept(): failed, "
						"continuing");
			}
			continue;
		}

		if (acl_check(cliaddr) == ACL_DENY) {
			logging_log(LOG_WARNING, "Access denied for host %s",
					sock_ntop(cliaddr));
			close(fd);
			continue;
		}

		pid = fork();
		switch (pid) {
		case 0:
			close(listenfd);
			log_client_address(cliaddr);
			serve_client(fd);
			logging_exit();
			_exit(0);
			/* NOTREACHED */
		case -1:
			logging_log(LOG_ERR, "fork(): failed, closing "
					"connection");
			close(fd);
			break;
		default:
			close(fd);
			break;
		}
	}

	/* NOTREACHED */
}
