/*
 * util.c -- common utility routines for thrulay.
 * 
 * Written by Stanislav Shalunov, http://www.internet2.edu/~shalunov/
 *            Bernhard Lutzmann, belu@users.sf.net
 * 
 * Copyright 2003, Internet2.
 * Legal conditions are in file LICENSE
 * (MD5 = ecfa50d1b0bfbb81b658c810d0476a52).
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include "util.h"
#include "rcs.h"

#ifndef SOL_IP
#ifdef IPPROTO_IP
#define SOL_IP		IPPROTO_IP
#endif
#endif

#ifndef SOL_IPV6
#ifdef IPPROTO_IPV6
#define SOL_IPV6	IPPROTO_IPV6
#endif
#endif

RCS_ID("@(#) $Id: util.c,v 1.1.1.1 2005/10/02 01:18:09 hliu Exp $")

void
error(int errcode, const char *msg)
{
	const char *prefix;
	int fatal = 1;

	if (errcode == ERR_FATAL) {
		prefix = "fatal";
	} else if (errcode == ERR_WARNING) {
		prefix = "warning";
		fatal = 0;
	} else {
		prefix = "UNKNOWN ERROR TYPE";
	}
	fprintf(stderr, "%s: %s\n", prefix, msg);
	if (fatal)
		exit(1);
}

ssize_t
read_exactly(int d, void *buf, size_t nbytes)
{
	ssize_t rc = 0;
	size_t bytes_read = 0;
	while (bytes_read < nbytes &&
	       (rc = read(d, (char *)buf+bytes_read,
			  nbytes-bytes_read)) > 0)
		bytes_read += rc;
	return rc == -1? rc: (ssize_t) bytes_read;
}

ssize_t
write_exactly(int d, const void *buf, size_t nbytes)
{
	ssize_t rc = 0;
	size_t bytes_written = 0;
	while (bytes_written < nbytes && 
	       (rc = write(d, (const char *)buf+bytes_written,
			   nbytes-bytes_written)) > 0)
		bytes_written += rc;
	return rc == -1? rc: (ssize_t) bytes_written;
}

double
time_diff(const struct timeval *tv1, const struct timeval *tv2)
{
	return (double) (tv2->tv_sec - tv1->tv_sec)
		+ (double) (tv2->tv_usec - tv1->tv_usec)/1000000.0;
}

/* Set window size for file descriptor FD (which must point to a
   socket) to WINDOW for given direction DIRECTION (typically
   SO_SNDBUF or SO_RCVBUF).  Try hard.  Return actual window size. */
int
set_window_size_directed(int fd, int window, int direction)
{
	int rc, try, w;
	unsigned int optlen = sizeof w;

	rc = getsockopt(fd, SOL_SOCKET, direction, (char *)&w, &optlen);
	if (rc == -1)
		return -1;
	if (window <= 0)
		return w;

	try = window;
	do {
		rc = setsockopt(fd, SOL_SOCKET, direction,
				(char *)&try, optlen);
		try *= 7;
		try /= 8;
	} while (try > w && rc == -1);

	rc = getsockopt(fd, SOL_SOCKET, direction, (char *)&w, &optlen);
	if (rc == -1)
		return -1;
	else
		return w;
}

/* Set window size for file descriptor FD (which must point to a
   socket) to WINDOW.  Try hard.  Return actual window size. */
int
set_window_size(int fd, int window)
{
	int send, receive;

	send = set_window_size_directed(fd, window, SO_SNDBUF);
	receive = set_window_size_directed(fd, window, SO_RCVBUF);
	return send < receive? send: receive;
}

/* Set RFC 2474 style DSCP value for TOS byte. Returns 0 on success, -1 on
 * failure. */
int
set_dscp(int sock, uint8_t dscp)
{
	int optname = IP_TOS;
	int optlevel = SOL_IP;
	int sopt;

	if ((dscp & ~0x3F)) {
		fprintf(stderr, "Error: set_dscp(): bad DSCP value.\n");
		return -1;
	}

	/* Shift to set CU to zero
	 *
	 *   0   1   2   3   4   5   6   7
	 * +---+---+---+---+---+---+---+---+
	 * |         DSCP          |  CU   |
	 * +---+---+---+---+---+---+---+---+
	 */
	sopt = dscp << 2;

	{
		union {
			struct sockaddr sa;
			char data[128];		/* max socket address
						   structure */
		} un;
		socklen_t len = 128;

		if (getsockname(sock, (struct sockaddr *)un.data, &len) == -1) {
			perror("getsockname");
			return -1;
		}

		switch (un.sa.sa_family) {
		case AF_INET:
			optlevel = SOL_IP;
			optname = IP_TOS;
			break;
		case AF_INET6:
#ifdef IPV6_TCLASS
			optlevel = SOL_IPV6;
			optname = IPV6_TCLASS;
			break;
#else
			error(ERR_WARNING, "system does not support setting "
					"DSCP value in IPv6 traffic class.");
			return 0;	/* return 0 so we don't get two
					   error messages. */
#endif
		default:
			error(ERR_WARNING, "set_dscp(): Unknown address "
					"family");
			return -1;
		}
	}

	if (setsockopt(sock, optlevel, optname, &sopt, sizeof(sopt)) == -1) {
		perror("setsockopt");
		return -1;
	}

	return 0;
}

#define NTP_EPOCH_OFFSET	2208988800ULL

/*
 * Convert `timeval' structure value into NTP format (RFC 1305) timestamp.
 * The ntp pointer must resolve to already allocated memory (8 bytes) that
 * will contain the result of the conversion.
 * NTP format is 4 octets of unsigned integer number of whole seconds since
 * NTP epoch, followed by 4 octets of unsigned integer number of
 * fractional seconds (both numbers are in network byte order).
 */
void
tv2ntp(const struct timeval *tv, char *ntp)
{
	uint32_t msb, lsb;

	msb = tv->tv_sec + NTP_EPOCH_OFFSET;
	lsb = (uint32_t)((double)tv->tv_usec * 4294967296.0 / 1000000.0);

	msb = htonl(msb);
	lsb = htonl(lsb);

	memcpy(ntp, &msb, sizeof(msb));
	memcpy(ntp + sizeof(msb), &lsb, sizeof(lsb));
}

/*
 * Convert 8-byte NTP format timestamp into `timeval' structure value.
 * The counterpart to tv2ntp().
 */
void
ntp2tv(struct timeval *tv, const char *ntp)
{
	uint32_t msb, lsb;

	memcpy(&msb, ntp, sizeof(msb));
	memcpy(&lsb, ntp + sizeof(msb), sizeof(lsb));

	msb = ntohl(msb);
	lsb = ntohl(lsb);

	tv->tv_sec = msb - NTP_EPOCH_OFFSET;
	tv->tv_usec = (uint32_t)((double)lsb * 1000000.0 / 4294967296.0);
}

/* Make sure 0 <= tv.tv_usec < 1000000.  Return 0 if it was normal,
 * positive number otherwise. */
int
normalize_tv(struct timeval *tv)
{
	int result = 0;

	while (tv->tv_usec >= 1000000) {
		tv->tv_usec -= 1000000;
		tv->tv_sec++;
		result++;
	}
	while (tv->tv_usec < 0) {
		tv->tv_usec += 1000000;
		tv->tv_sec--;
		result++;
	}
	return result;
}
